import numpy as np
import os, time, pickle
import napari
from qtpy.QtWidgets import QPushButton, QVBoxLayout, QTabWidget, QWidget
from napari.utils import progress
from magicgui.widgets import TextEdit
from skimage.segmentation import find_boundaries, watershed
from skimage.measure import label
from skimage.morphology import skeletonize, binary_closing, binary_opening
from skimage.util import apply_parallel
from skimage.measure import regionprops, regionprops_table
from multiprocessing.pool import ThreadPool as Pool
import pandas as pand

import epicure.Utils as ut
from epicure.editing import Editing
from epicure.tracking import Tracking
from epicure.inspecting import Inspecting
from epicure.outputing import Outputing
from epicure.displaying import Displaying
from epicure.preferences import Preferences

"""
    EpiCure main
    Open and initialize the files
    Launch the main widget composed of the segmentation and tracking editing features
"""

class EpiCure():
    
    def __init__(self, viewer=None):
        self.viewer = viewer
        if self.viewer is None:
            self.viewer = napari.Viewer(show=False)
        self.viewer.title = "Napari - EpiCure"
        self.img = None
        self.inspecting = None
        self.others = None
        self.imgshape2D = None    ## width, height of the image
        self.nframes = None       ## Number of time frames
        self.thickness = 4 ## thickness of junctions, wider
        self.minsize = 4   ## smallest number of pixels in a cell
        self.verbose = 1  ## level of printing messages (None/few, normal, debug mode)
        
        self.overtext = dict()
        self.help_index = 1   ## current display index of help overlay
        self.blabla = None    ## help window
        self.groups = {}
        self.tracked = 0 ## has done a tracking
        self.process_parallel = False ## Do some operations in parallel (n frames in parallel)
        self.nparallel = 4      ## number of parallel threads
        self.dtype = np.uint32  ## label type, default 32 but if less labels, reduce it
        self.outputing = None   ## non initialized yet
        
        self.forbid_gaps = False  ## allow gaps in track or not

        self.pref = Preferences()
        self.shortcuts = self.pref.get_shortcuts() ## user specific shortcuts
        self.settings = self.pref.get_settings() ## user specific preferences
        ## display settings
        self.display_colors = None  ## settings for changing some display colors
        if "Display" in self.settings:
            if "Colors" in self.settings["Display"]:
                self.display_colors = self.settings["Display"]["Colors"]
        
        self.init_epicure_metadata()   ## initialize metadata variables (scalings, channels)

    def init_epicure_metadata( self ):
        """ Returns metadata to save """
        ## scalings and unit names
        self.epi_metadata = {}
        self.epi_metadata["ScaleXY"] = 1
        self.epi_metadata["UnitXY"] = "um"
        self.epi_metadata["ScaleT"] = 1
        self.epi_metadata["UnitT"] = "min"
        self.epi_metadata["MainChannel"] = 0

    def get_resetbtn_color( self ):
        """ Returns the color of Reset buttons if defined """
        if "Display" in self.settings:
            if "Colors" in self.settings["Display"]:
                if "Reset button" in self.settings["Display"]["Colors"]:
                    return self.settings["Display"]["Colors"]["Reset button"]
        return None

    def set_thickness( self, thick ):
        """ Thickness of junctions (half thickness) """
        self.thickness = thick

    def load_movie(self, imgpath):
        """ Load the intensity movie, and get metadata """
        self.imgpath = imgpath
        self.img, nchan, self.epi_metadata["ScaleXY"], self.epi_metadata["UnitXY"], self.epi_metadata["ScaleT"], self.epi_metadata["UnitT"] = ut.opentif( self.imgpath, verbose=self.verbose>1 )
        ## transform static image to movie (add temporal dimension)
        if len(self.img.shape) == 2:
            self.img = np.expand_dims(self.img, axis=0)
        caxis = None
        cval = 0
        if nchan>0 or len(self.img.shape) > 3:
            if nchan > 0 and len(self.img.shape) > 3:
                ## multiple chanels and multiple slices, order axis should be TCXY
                caxis = 1
                cval = nchan
            else:
                ## image has multiple chanels
                minshape = min(self.img.shape)
                caxis = self.img.shape.index(minshape)
                cval = minshape
            self.mov = self.img

        ## display the movie
        ut.remove_layer(self.viewer, "Movie")
        mview = self.viewer.add_image( self.img, name="Movie", blending="additive", colormap="gray" )
        mview.contrast_limits = self.quantiles()
        mview.gamma = 0.95
        
        self.imgshape = self.viewer.layers["Movie"].data.shape
        self.imgshape2D = self.imgshape[1:3]
        self.nframes = self.imgshape[0]
        return caxis, cval

    def quantiles(self):
        return tuple(np.quantile(self.img, [0.01, 0.9999]))

    def set_scales( self, scalexy, scalet, unitxy, unitt ):
        """ Set the scaling units for outputs """
        self.epi_metadata["ScaleXY"] = scalexy
        self.epi_metadata["ScaleT"] = scalet
        self.epi_metadata["UnitXY"] = unitxy
        self.epi_metadata["UnitT"] = unitt
        if self.verbose > 0:
            ut.show_info( "Movie scales set to "+str(self.epi_metadata["ScaleXY"])+" "+self.epi_metadata["UnitXY"]+" and "+str(self.epi_metadata["ScaleT"])+" "+self.epi_metadata["UnitT"] )

    def set_chanel( self, chan, chanaxis ):
        """ Update the movie to the correct chanel """
        self.img = np.rollaxis(np.copy(self.mov), chanaxis, 0)[chan]
        self.main_channel = chan
        if self.viewer is not None:
            mview = self.viewer.layers["Movie"]
            mview.data = self.img
            mview.contrast_limits=self.quantiles()
            mview.gamma=0.95
            mview.refresh()

    def add_other_chanels(self, chan, chanaxis): 
        """ Open other channels if option selected """
        self.others = np.delete(self.mov, chan, axis=chanaxis)
        self.others_chanlist = []
        if self.others is not None:
            self.others = np.rollaxis(self.others, chanaxis, 0)
            for ochan in range(self.others.shape[0]):
                purechan = ochan
                if purechan >= chan:
                    purechan = purechan + 1
                self.others_chanlist.append(purechan)
                mview = self.viewer.add_image( self.others[ochan], name="MovieChannel_"+str(purechan), blending="additive", colormap="gray" )
                mview.contrast_limits=tuple(np.quantile(self.others[ochan],[0.01, 0.9999]))
                mview.gamma=0.95
                mview.visible = False
        
    def load_segmentation(self, segpath):
        """ Load the segmentation file """
        start_time = ut.start_time()
        self.segpath = segpath
        self.seg,_, _,_,_,_ = ut.opentif( self.segpath, verbose=self.verbose>1 )
        self.seg = np.uint32(self.seg)
        ## transform static image to movie (add temporal dimension)
        if len(self.seg.shape) == 2:
            self.seg = np.expand_dims(self.seg, axis=0)
        ## ensure that the shapes are correctly set
        self.imgshape = self.seg.shape
        self.imgshape2D = self.seg.shape[1:3]
        self.nframes = self.seg.shape[0]
        ## if the segmentation is a junction file, transform it to a label image 
        if ut.is_binary(self.seg):
            self.junctions_to_label()
            self.tracked = 0
        else:
            self.has_been_tracked()
            self.thin_boundaries()

        if np.max(self.seg) < 50000:
            self.dtype = np.uint16
            self.seg = np.uint16(self.seg)
        
        # display the segmentation file movie
        if self.viewer is not None:
            self.seglayer = self.viewer.add_labels( self.seg, name="Segmentation", blending="additive", opacity=0.5 )
            self.viewer.dims.set_point(0,0)
            self.seglayer.brush_size = 4 ## default label pencil drawing size 
        if self.verbose > 0:
            ut.show_duration(start_time, header="Segmentation loaded in ")

    def load_tracks( self, progress_bar ):
        """ From the segmentation, get all the metadata """
        tracked = "tracked"
        self.tracking.init_tracks()
        if self.tracked == 0:
            tracked = "untracked"
        else:
            if self.forbid_gaps:
                progress_bar.set_description( "check and fix track gaps" )
                self.handle_gaps( track_list=None, verbose=1 )
        ut.show_info(""+str(len(self.tracking.get_track_list()))+" "+tracked+" cells loaded")


    def has_been_tracked(self):
        """ Look if has been tracked already (some labels are in several frames) """
        nb = 0
        for frame in range(self.seg.shape[0]):
            if frame > 0:
                inter = np.intersect1d(np.unique(self.seg[frame-1]), np.unique(self.seg[frame]))
                if len(inter)>1:
                    self.tracked = 1 
                    return
        self.tracked = 0
        return

    def suggest_segfile(self, outdir):
        """ Check if a segmentation file from EpiCure already exists """
        imgname, imgdir, out = ut.extract_names( self.imgpath, outdir, mkdir=False )
        return ut.suggest_segfile(out, imgname)

    def outname(self):
        return os.path.join(self.outdir, self.imgname)

    def set_names(self, outdir):
        """ Extract default names from imgpath """
        self.imgname, self.imgdir, self.outdir = ut.extract_names( self.imgpath, outdir, mkdir=True )

    def go_epicure( self, outdir="epics", segmentation_file=None ):
        """ Initialize everything and start the main widget """
        self.set_names( outdir )
        if segmentation_file is None:
            segmentation_file = self.suggest_segfile( outdir )
        self.viewer.window._status_bar._toggle_activity_dock(True)
        progress_bar = progress(total=5)
        progress_bar.set_description( "Reading segmented image" )
        ## load the segmentation
        self.load_segmentation( segmentation_file )
        progress_bar.update(1)
        ut.set_active_layer( self.viewer, "Segmentation" )
        
        ## setup the main interface and shortcuts
        start_time = ut.start_time()
        progress_bar.set_description( "Active EpiCure shortcuts" )
        self.key_bindings()
        progress_bar.update(2)
        progress_bar.set_description( "Prepare widget" )
        self.main_widget()
        progress_bar.update(3)
        progress_bar.set_description( "Load tracks" )
        self.load_tracks( progress_bar )
        progress_bar.update(4)

        ## load graph if it exists
        epiname = os.path.join( self.outdir, self.imgname+"_epidata.pkl" )
        if os.path.exists(epiname):
            progress_bar.set_description( "Load EpiCure informations" )
            self.load_epicure_data(epiname)
        if self.verbose > 0:
            ut.show_duration(start_time, header="Tracks and graph loaded in ")
        progress_bar.update(5)
        self.apply_settings()
        progress_bar.close()
        self.viewer.window._status_bar._toggle_activity_dock(False)

###### Settings (preferences) save and load
    def apply_settings( self ):
        """ Apply all default or prefered settings """
        for sety, val in self.settings.items():
            if sety=="Display":
                self.display.apply_settings( val )
                if "Show help" in val:
                    index = int( val["Show help"] )
                    self.switchOverlayText( index )
                if "Contour" in val:
                    contour = int( val["Contour"] )
                    self.seglayer.contour = contour
                    self.seglayer.refresh()
                if "Colors" in val:
                    color = val["Colors"]["button"]
                    check_color = val["Colors"]["checkbox"]
                    line_edit_color = val["Colors"]["line edit"]
                    group_color = val["Colors"]["group"]
                    self.main_gui.setStyleSheet( 'QPushButton {background-color: '+color+'} QCheckBox::indicator {background-color: '+check_color+'} QLineEdit {background-color: '+line_edit_color+'} QGroupBox {color: grey; background-color: '+group_color+'} ')
                    self.display_colors = val["Colors"]
            if sety == "events":
                self.inspecting.apply_settings( val )
            if sety == "Output":
                self.outputing.apply_settings( val )
            if sety == "Track":
                self.tracking.apply_settings( val )
            if sety == "Edit":
                self.editing.apply_settings( val )
            #case _:
            #       continue
            ## match is not compatible with python 3.9

    def update_settings( self ):
        """ Returns all the prefered settings """
        disp = self.settings
        ## load display current settings (layers visibility)
        disp["Display"] = self.display.get_current_settings()
        disp["Display"]["Show help"] = self.help_index
        disp["Display"]["Contour"] = self.seglayer.contour
        ## load suspect current settings
        disp["events"] = self.inspecting.get_current_settings()
        ## get outputs current settings
        disp["Output"] = self.outputing.get_current_settings()
        disp["Track"] = self.tracking.get_current_settings()
        disp["Edit"] = self.editing.get_current_settings()

#### Main widget that contains the tabs of the sub widgets
    
    def main_widget(self):
        """ Open the main widget interface """
        self.main_gui = QWidget()
        
        layout = QVBoxLayout()
        tabs = QTabWidget()
        tabs.setObjectName("main")
        layout.addWidget(tabs)
        self.main_gui.setLayout(layout)
        
        self.editing = Editing(self.viewer, self)
        tabs.addTab( self.editing, "Edit" )
        self.inspecting = Inspecting(self.viewer, self)
        tabs.addTab( self.inspecting, "Inspect" )
        self.tracking = Tracking(self.viewer, self)
        tabs.addTab( self.tracking, "Track" )
        self.outputing = Outputing(self.viewer, self)
        tabs.addTab( self.outputing, "Output" )
        self.display = Displaying(self.viewer, self)
        tabs.addTab( self.display, "Display" )
        self.main_gui.setStyleSheet('QPushButton {background-color: rgb(40, 60, 75)} QCheckBox::indicator {background-color: rgb(40,52,65)}')

        self.viewer.window.add_dock_widget( self.main_gui, name="Main" )

    def key_bindings(self):
        """ Activate shortcuts """
        self.text = "-------------- ShortCuts -------------- \n "
        self.text += "!! Shortcuts work if Segmentation layer is active !! \n"
        #for sctype, scvals in self.shortcuts.items():
        self.text += "\n---"+"General"+" options---\n"
        sg = self.shortcuts["General"]
        self.text += ut.print_shortcuts( sg )
        self.text = self.text + "\n"
        
        if self.verbose > 0:
            print("Activating key shortcuts on segmentation layer")
            print("Press <" + str(sg["show help"]["key"]) + "> to show/hide the main shortcuts")
            print("Press <" + str(sg["show all"]["key"]) + "> to show ALL shortcuts")
        ut.setOverlayText(self.viewer, self.text, size=10)
        
        @self.seglayer.bind_key( sg["show help"]["key"], overwrite=True )
        def switch_shortcuts(seglayer):
            #index = (self.help_index+1)%(len(self.overtext.keys())+1)
            #self.switchOverlayText(index)
            index = (self.help_index+1)%2
            self.switchOverlayText( index )
        
        @self.seglayer.bind_key( sg["show all"]["key"], overwrite=True )
        def list_all_shortcuts(seglayer):
            self.switchOverlayText(0)   ## hide display message in main window
            text = "**************** EPICURE *********************** \n"
            text += "\n"
            text += self.text
            text += "\n"
            text += ut.napari_shortcuts()
            for key, val in self.overtext.items():
                text += "\n"
                text += val
            self.update_text_window(text) 
        
        @self.seglayer.bind_key( sg["save segmentation"]["key"], overwrite=True )
        def save_seglayer(seglayer):
            self.save_epicures()
        
        @self.viewer.bind_key( sg["save movie"]["key"], overwrite=True )
        def save_movie(seglayer):
            endname = "_frames.tif"
            outname = os.path.join( self.outdir, self.imgname+endname )
            self.save_movie(outname)

########### Texts 

    def switchOverlayText(self, index):
        """ Switch overlay display text to index """
        self.help_index = index
        if index == 0:
            ut.showOverlayText(self.viewer, vis=False)
            return
        else:
            ut.showOverlayText(self.viewer, vis=True)
        #self.setCurrentOverlayText()
        self.setGeneralOverlayText()

    def init_text_window(self):
        """ Create and display help text window """
        self.blabla = TextEdit()
        self.blabla.name = "Epicure shortcuts"
        self.blabla.show()

    def update_text_window(self, message):
        """ Update message in separate window """
        self.init_text_window()
        self.blabla.value = message

    def setGeneralOverlayText(self):
        """ set overlay help message to general message """
        text = self.text
        ut.setOverlayText(self.viewer, text, size=10)

    def setCurrentOverlayText(self):
        """ Set overlay help text message to current selected options list """
        text = self.text
        dispkey = list(self.overtext.keys())[self.help_index-1]
        text += self.overtext[dispkey]
        ut.setOverlayText(self.viewer, text, size=10)
    
    def get_summary(self):
        """ Get a summary of the infos of the movie """
        summ = "----------- EpiCure summary ----------- \n"
        summ += "--- Image infos \n"
        summ += "Movie name: "+str(self.imgpath)+"\n"
        summ += "Movie size (x,y): "+str(self.imgshape2D)+"\n"
        if self.nframes is not None:
            summ += "Nb frames: "+str(self.nframes)+"\n"
        summ += "\n"
        summ += "--- Segmentation infos \n"
        summ += "Segmentation file: "+str(self.segpath)+"\n"
        summ += "Nb tracks: "+str(len(self.tracking.get_track_list()))+"\n"
        tracked = "yes"
        if self.tracked == 0:
            tracked = "no"
        summ += "Tracked: "+tracked+"\n"
        nb_labels, mean_duration, mean_area = ut.summary_labels( self.seg )
        summ += "Nb cells: "+str( nb_labels )+"\n"
        summ += "Average track lengths: "+str(mean_duration)+" frames\n"
        summ += "Average cell area: "+str(mean_area)+" pixels^2\n"
        summ += "Nb suspect events: "+str(self.inspecting.nb_events(only_suspect=True))+"\n"
        summ += "Nb divisions: "+str(self.inspecting.nb_type("division"))+"\n"
        summ += "\n"
        summ += "--- Parameter infos \n"
        summ += "Junction thickness: "+str(self.thickness)+"\n"
        return summ

    def set_contour( self, width ):
        self.seglayer.contour = width

############ Layers

    def check_layers(self):
        """ Check that the necessary layers are present """
        if self.editing.shapelayer_name not in self.viewer.layers:
            if self.verbose > 0:
                print("Reput shape layer")
            self.editing.create_shapelayer()
        if self.inspecting.eventlayer_name not in self.viewer.layers:
            if self.verbose > 0:
                print("Reput event layer")
            self.inspecting.create_eventlayer()
        if "Movie" not in self.viewer.layers:
            if self.verbose > 0:
                print("Reput movie layer")
            mview = self.viewer.add_image( self.img, name="Movie", blending="additive", colormap="gray" )
            #mview.reset_contrast_limits()
            mview.contrast_limits=self.quantiles()
            mview.gamma=0.95
        if "Segmentation" not in self.viewer.layers:
            if self.verbose > 0:
                print("Reput segmentation")
            self.seglayer = self.viewer.add_labels( self.seg, name="Segmentation", blending="additive", opacity=0.5 )

        self.finish_update()


    def finish_update(self, contour=None):
        if contour is not None:
            self.seglayer.contour = contour
        ut.set_active_layer( self.viewer, "Segmentation" )
        self.seglayer.refresh()
        duplayers = ["PrevSegmentation"]
        for dlay in duplayers:
            if dlay in self.viewer.layers:
                (self.viewer.layers[dlay]).refresh()
    
    def read_epicure_metadata( self ):
        """ Load saved infos from file """
        epiname = self.outname() + "_epidata.pkl" 
        if os.path.exists( epiname ):
            infile = open(epiname, "rb")
            try:
                epidata = pickle.load( infile )
                if "EpiMetaData" in epidata.keys():
                    self.epi_metadata = epidata["EpiMetaData"]
                infile.close()
            except:
                ut.show_warning( "Could not read EpiCure data file "+epiname )


    def save_epicures( self, imtype="float32" ):
        outname = os.path.join( self.outdir, self.imgname+"_labels.tif" )
        ut.writeTif(self.seg, outname, self.epi_metadata["ScaleXY"], imtype, what="Segmentation")
        epiname = os.path.join( self.outdir, self.imgname+"_epidata.pkl" )
        outfile = open(epiname, "wb")
        epidata = {}
        epidata["EpiMetaData"] = self.epi_metadata
        if self.groups is not None:
            epidata["Group"] = self.groups
        if self.tracking.graph is not None:
            epidata["Graph"] = self.tracking.graph
        if self.inspecting is not None and self.inspecting.events is not None:
            epidata["Events"] = {}
            if self.inspecting.events.data is not None:
                epidata["Events"]["Points"] = self.inspecting.events.data
                epidata["Events"]["Props"] = self.inspecting.events.properties
                epidata["Events"]["Types"] = self.inspecting.event_types
                epidata["Events"]["Symbols"] = self.inspecting.events.symbol
                epidata["Events"]["Colors"] = self.inspecting.events.face_color
        pickle.dump( epidata, outfile )
        outfile.close()
    
    def read_group_data( self, groups ):
        """ Read the group EpiCure data from opened file """
        if self.verbose > 0:
            print( "Loaded cell groups info: "+str(list(groups.keys())) )
            if self.verbose > 2:
                print( "Cell groups: "+str(groups) )
        return groups
        
    def read_graph_data( self, infile ):
        """ Read the graph EpiCure data from opened file """
        try:
            graph = pickle.load(infile)
            if self.verbose > 0:
                print("Graph (lineage) loaded")
            return graph
        except:
            if self.verbose > 1:
                print("No graph infos found")
            return None

    def read_events_data(self, infile):
        """ Read info of EpiCure events (suspects, divisions) from opened file """
        try:
            events_pts = pickle.load(infile)
            if events_pts is not None:
                events_props = pickle.load(infile)
                events_type = pickle.load(infile)
                try:
                    symbols = pickle.load(infile)
                    colors = pickle.load(infile)
                except: 
                    if self.verbose > 1:
                        print("No events display info found")
                    symbols = None
                    colors = None
                return events_pts, events_props, events_type, symbols, colors
            else:
                return None, None, None, None, None
        except:
            if self.verbose > 1:
                print("events info not complete")
            return None, None, None, None, None

    def load_epicure_data(self, epiname):
        """ Load saved infos from file """
        infile = open(epiname, "rb")
        try:
            epidata = pickle.load( infile )
            if "EpiMetaData" in epidata.keys():
                # version of epicure file after Epicure 0.2.0
                self.read_epidata( epidata )
                infile.close()
            else:
                # version anterior of Epicure 0.2.0
                self.load_epicure_data_old( epidata, infile )
        except:
            ut.show_warning( "Could not read EpiCure data file "+epiname )

    def read_epidata( self, epidata ):
        """ Read the dict of saved state and initialize all instances with it """
        for key, vals in epidata.items():
            if key == "EpiMetaData":
                ## image data is read on the previous step
                continue
            if key == "Group":
                ## Load groups information
                self.groups = self.read_group_data( vals )
                for group in self.groups.keys():
                    self.editing.update_group_list( group )
                self.outputing.update_selection_list()
            if key == "Graph":
                ## Load graph (lineage) informations
                self.tracking.graph = vals
                if self.tracking.graph is not None:
                    self.tracking.tracklayer.refresh()
            if key == "Events":
                ## Load events information
                if "Points" in vals.keys():
                    pts = vals["Points"]
                if "Props" in vals.keys():
                    props = vals["Props"]
                if "Types" in vals.keys():
                    event_types = vals["Types"]
                if "Symbols" in vals.keys():
                    symbols = vals["Symbols"]
                if "Colors" in vals.keys():
                    colors = vals["Colors"] 
                if pts is not None:
                    if len(pts) > 0:
                        self.inspecting.load_events(pts, props, event_types, symbols, colors)
                    if len(pts) > 0 and self.verbose > 0:
                        print("events loaded")
                    ut.show_info("Loaded "+str(len(pts))+" events")     
               

    def load_epicure_data_old( self, groups, infile ):
        """ Load saved infos from file """
        ## Load groups information
        self.groups = self.read_group_data( groups )
        for group in self.groups.keys():
            self.editing.update_group_list( group )
        self.outputing.update_selection_list()
        ## Load graph (lineage) informations
        self.tracking.graph = self.read_graph_data( infile )
        if self.tracking.graph is not None:
            self.tracking.tracklayer.refresh()
        ## Load events information 
        pts, props, event_types, symbols, colors = self.read_events_data( infile )
        if pts is not None:
            if len(pts) > 0:
                self.inspecting.load_events(pts, props, event_types, symbols, colors)
                if len(pts) > 0 and self.verbose > 0:
                    print("events loaded")
                    ut.show_info("Loaded "+str(len(pts))+" events")
        infile.close()

    def save_movie(self, outname):
        """ Save movie with current display parameters, except zoom """
        save_view = self.viewer.camera.copy()
        save_frame = ut.current_frame(self.viewer)
        ## place the view to see the whole image
        self.viewer.reset_view()
        #self.viewer.camera.zoom = 1
        sizex = (self.imgshape2D[0]*self.viewer.camera.zoom)/2
        sizey = (self.imgshape2D[1]*self.viewer.camera.zoom)/2
        if os.path.exists(outname):
            os.remove(outname)

        ## take a screenshot of each frame
        for frame in range(self.nframes):
            self.viewer.dims.set_point(0, frame)
            shot = self.viewer.window.screenshot(canvas_only=True, flash=False)
            ## remove border: movie is at the center
            centx = int(shot.shape[0]/2)+1
            centy = int(shot.shape[1]/2)+1
            shot = shot[int(centx-sizex):int(centx+sizex), int(centy-sizey):int(centy+sizey),]
            ut.appendToTif(shot, outname)
        self.viewer.camera.update(save_view)
        if save_frame is not None:
            self.viewer.dims.set_point(0, save_frame)
        ut.show_info("Movie "+outname+" saved")

    def reset_data( self ):
        """ Reset EpiCure data (group, suspect, graph) """
        self.inspecting.reset_all_events()
        self.reset_groups()
        self.outputing.update_selection_list()
        self.tracking.graph = None

    def junctions_to_label(self):
        """ convert epyseg/skeleton result (junctions) to labels map """
        cmax = 0
        for z in range(self.seg.shape[0]):
            self.label_one_frame(z)
            ## shift all the labels with the previous max
            if z > 0:
                nonzero = self.seg[z] > 0
                self.seg[z,nonzero] = self.seg[z,nonzero] + cmax
            cmax = np.max(self.seg[z])

    def shift_labels(self):
        """ Shift the labels of each frame so they never overlap """
        cmax = 0
        for z in range(self.seg.shape[0]):
            if z > 0:
                nonzero = self.seg[z] > 0
                self.seg[z,nonzero] = self.seg[z,nonzero] + cmax
            cmax = np.max(self.seg[z])

    def label_one_frame(self, z):
        """ From segmentation of junctions of one frame, get it as a label frame """
        skel = skeletonize( self.seg[z]/np.max(self.seg[z] ))
        skel = self.copy_border( skel, self.seg[z] )
        self.seg[z] = label( skel, background=1, connectivity=1)
        return

    def copy_border( self, skel, bin ):
        """ Copy the pixel border onto skeleton image """
        skel[0,:] = bin[0,:] ## borders
        skel[:,0] = bin[:,0]
        skel[skel.shape[0]-1,:] = bin[skel.shape[0]-1,:]
        skel[:,skel.shape[1]-1] = bin[:,skel.shape[1]-1]
        return skel
        
    def thin_boundaries(self):
        """" Assure that all boundaries are only 1 pixel thick """
        if self.process_parallel:
            pool = Pool(self.nparallel)
            pool.map( self.thin_seg_one_frame, range(self.seg.shape[0]) )
            pool.close()
        else:
            for z in range(self.seg.shape[0]):
                self.thin_seg_one_frame(z)

    def skeleton_to_label( self, skel, labelled ):
        """ Transform a skeleton to label image with numbers from labelled image """
        newlab = label( skel, background=np.max(skel), connectivity=1 )
        props = regionprops( newlab )  
        newlab = np.zeros( newlab.shape, np.uint32 )   
        for prop in props:
            if prop.label != 0:
                labvals, counts = np.unique(labelled[prop.bbox[0]:prop.bbox[2],prop.bbox[1]:prop.bbox[3]][prop.image], return_counts=True )
                labval = labvals[ np.argmax(counts) ]
                newlab[prop.bbox[0]:prop.bbox[2],prop.bbox[1]:prop.bbox[3]][prop.image] = labval
        return newlab

    def thin_seg_one_frame(self, tframe):
        """ Boundaries of the frame one pixel thick """
        bin_img = binary_closing( find_boundaries(self.seg[tframe], connectivity=2, mode="outer"), footprint=np.ones((3,3)) )
        skel = skeletonize( bin_img )
        skel = self.copy_border( skel, bin_img )
        self.seg[tframe] = self.skeleton_to_label( skel, self.seg[tframe] )

    def frame_to_skeleton(self, frame, connectivity=1):
        """ convert labels frame to skeleton (thin boundaries) """
        return skeletonize( find_boundaries(frame, connectivity=connectivity, mode="outer") )

    def add_skeleton(self):
        """ add a layer containing the skeleton movie of the segmentation """
        # display the segmentation file movie
        if self.viewer is not None:
            skel = np.zeros(self.seg.shape, dtype="uint8")
            skel[self.seg==0] = 1
            skel = self.get_skeleton()
            ut.remove_layer(self.viewer, "Skeleton")
            skellayer = self.viewer.add_image( skel, name="Skeleton", blending="additive", opacity=1 )
            skellayer.reset_contrast_limits()
            skellayer.contrast_limits = (0,1)
    
    def get_skeleton(self):
        """ convert labels movie to skeleton (thin boundaries) """
        if self.seg is None:
            return None
        return ut.get_skeleton( self.seg, verbose = self.verbose )

    def to_skeleton(self, mode):
        """ convert labels movie to skeleton (thin boundaries) """
        if self.seg is not None:
            start_time = ut.start_time()
            skel = np.zeros(self.seg.shape, dtype="uint8")
            for z in range(self.seg.shape[0]):
                skel[z,] = apply_parallel( self.frame_to_skeleton, self.seg[z,], depth=5, compute=True )
        if self.verbose > 0:
            ut.show_duration(start_time)
        return skel


    ############ Label functions

    def get_free_labels(self, nlab):
        """ Get the nlab smallest unused labels """
        used = set(self.tracking.get_track_list())
        return ut.get_free_labels( used, nlab )

    def get_free_label(self):
        """ Return the first free label """
        return self.get_free_labels(1)[0]

    def has_label(self, label):
        """ Check if label is present in the tracks """
        return self.tracking.has_track(label)
    
    def has_labels(self, labels):
        """ Check if labels are present in the tracks """
        return self.tracking.has_tracks( labels )

    def nlabels(self):
        """ Number of unique tracks """
        return self.tracking.nb_tracks()

    def get_labels(self):
        """ Return list of labels in tracks """
        return list(self.tracking.get_track_list())

    ########## Edit tracks
    def delete_tracks(self, tracks):
        """ Remove all the tracks from the Track layer """
        self.tracking.remove_tracks( tracks )

    def delete_track(self, label, frame=None):
        """ Remove (part of) the track """
        if frame is None:
            self.tracking.remove_track(label)
        else:
            self.tracking.remove_one_frame(label, frame, handle_gaps=self.forbid_gaps )

    def update_centroid(self, label, frame):
        """ Track label has been change at given frame """
        if label not in self.tracking.has_track(label):
            if self.verbose > 1:
                print("Track "+str(label)+" not found")
            return
        self.tracking.update_centroid(label, frame)

    ########## Edit label
    def get_label_indexes (self, label, start_frame=0 ):
        """ Returns the indexes where label is present in segmentation, starting from start_frame """
        indmodif = []
        while start_frame < self.nframes:
            found_label = np.argwhere( self.seg[start_frame] == label )
            if found_label is None:
                break   ### if lose a label, stop there (don't allow gap frame)
            inds = ut.shiftFrameIndices( found_label.tolist(), start_frame )
            indmodif = indmodif + inds
            start_frame = start_frame + 1
        return indmodif

    def replace_label( self, label, new_label, start_frame=0 ):
        """ Replace label with new_label from start_frame """
        indmodif = self.get_label_indexes( label, start_frame )
        new_labels = [new_label]*len(indmodif)
        self.change_labels( indmodif, new_labels )
    
    def change_labels(self, indmodif, new_labels):
        """ Change the value at pixels indmodif to new_labels and update tracks/graph """
        if len(indmodif)>0:
            bbox = ut.getBBoxFromPts( indmodif, extend=0, imshape=self.imgshape, outdim=3 )
            ## get effectively changed labels
            indmodif, new_labels, old_labels = ut.setNewLabel( self.seglayer, indmodif, new_labels, add_frame=None )
            if len(new_labels) > 0:
                self.update_changed_labels( indmodif, new_labels, old_labels )
        self.seglayer.refresh()

    def get_mask( self, label, start=None, end=None ):
        """ Get mask of label from frame start to frame end """
        if (start is None) or (end is None):
            start, end = self.tracking.get_extreme_frames( label )
        crop = self.seg[start:(end+1)]
        mask = np.isin( crop, [label] )*1
        return mask

    def get_label_movie( self, label, extend=1.25 ):
        """ Get movie centered on label """
        start, end = self.tracking.get_extreme_frames( label )
        mask = self.get_mask( label, start, end )
        boxes = []
        centers = []
        max_box = 0
        for frame in mask:
            props = regionprops( frame )
            bbox = props[0].bbox
            boxes.append(bbox)
            centers.append(props[0].centroid)
            for i in range(2):
                max_box = max( max_box, bbox[i+2]-bbox[i] )
        
        box_size = int(max_box*extend)
        movie = np.zeros((end-start+1, box_size, box_size))
        for i, frame in enumerate(range(start, end+1)):
            xmin = int(centers[i][0] - box_size/2)
            xminshift=0
            if xmin < 0:
                xminshift = -xmin
                xmin = 0
            xmax = xmin + box_size - xminshift
            xmaxshift = box_size
            if xmax > self.imgshape2D[0]:
                xmaxshift = self.imgshape2D[0] - xmax
                xmax = self.imgshape2D[0]

            ymin = int(centers[i][1] - max_box/2)
            yminshift=0
            if ymin < 0:
                yminshift = -ymin
                ymin = 0
            ymax = ymin + box_size - yminshift
            ymaxshift = box_size
            if ymax > self.imgshape2D[1]:
                ymaxshift = self.imgshape2D[1] - ymax
                ymax = self.imgshape2D[1]

            movie[ i, xminshift:xmaxshift, yminshift:ymaxshift ] = self.img[ frame, xmin:xmax, ymin:ymax ]
        return movie 

    ### Check individual cell features
    def cell_radius( self, label, frame ):
        """ Approximate the cell radius at given frame """
        area = np.sum( self.seg[frame] == label )
        radius = math.sqrt(area/math.pi)
        return radius
    
    def cell_area( self, label, frame ):
        """ Approximate the cell radius at given frame """
        area = np.sum( self.seg[frame] == label )
        return area 

    def cell_on_border( self, label, frame ):
        """ Check if a given cell is on border of the image """
        bbox = ut.getBBox2D( self.seg[frame], label )
        out = ut.outerBBox2D( bbox, self.imgshape2D, margin=3 )
        return out

    ###### Synchronize tracks whith labels changed
    def add_label( self, labels, frame=None ):
        """ Add a label to the tracks """
        if frame is not None:
            if np.isscalar(labels):
                labels = [labels]
            self.tracking.add_one_frame( labels, frame, refresh=True )
        else:
            if self.verbose > 1:
                print("TODO add label no frame")
    
    def add_one_label_to_track( self, label ):
        """ Add the track data of a given label if missing """
        iframe = 0
        while (iframe < self.nframes) and (label not in self.seg[iframe]):
            iframe = iframe + 1
        while (iframe < self.nframes) and (label in self.seg[iframe]):
            self.tracking.add_one_frame( [label], iframe )
            iframe = iframe + 1

    def update_label(self, label, frame):
        """ Update the given label at given frame """
        self.tracking.update_track_on_frame([label], frame)

    def update_changed_labels( self, indmodif, new_labels, old_labels ):
        """ Check what had been modified, and update tracks from it """
        ## check all the old_labels if still present or not
        if self.verbose > 1:
            start_time = time.time()
        frames = np.sort( np.unique( indmodif[0] ) )
        all_deleted = []
        debug_verb = self.verbose > 2
        if debug_verb:
            print( "Updating labels in frames "+str(frames) )
        for frame in frames:
            keep = indmodif[0] == frame
            ## check old labels if totally removed or not
            deleted = np.setdiff1d( old_labels[keep], self.seg[frame] )
            if deleted.shape[0] > 0:
                self.tracking.remove_one_frame( deleted, frame, handle_gaps=False, refresh=False )
                all_deleted = all_deleted + list(set(deleted) - set(all_deleted))
            ## now check new labels
            nlabels = np.unique( new_labels[keep] )
            if nlabels.shape[0] > 0:
                self.tracking.update_track_on_frame( nlabels, frame )
            if debug_verb:
                print("Labels deleted at frame "+str(frame)+" "+str(deleted)+" or added "+str(nlabels))

        ## Check if some gaps has been created in tracks (remove middle(s) frame(s))
        if self.forbid_gaps:
            if len(all_deleted) > 0:
                self.handle_gaps( all_deleted, verbose=0 )

        if self.verbose > 1:
            ut.show_duration(start_time, "updated tracks in ")

    def handle_gaps(self, track_list, verbose=None):
        """ Check and fix gaps in tracks """
        if verbose is None:
            verbose = self.verbose
        gaped = self.tracking.check_gap( track_list, verbose=verbose )
        if len(gaped) > 0:
            if self.verbose > 0:
                print("Relabelling tracks with gaps")
            self.fix_gaps(gaped)
            
    def fix_gaps(self, gaps):
        """ Fix when some gaps has been created in tracks """
        for gap in gaps:
            gap_frames = self.tracking.gap_frames( gap )
            cur_gap = gap
            for gapy in gap_frames:
                new_value = self.get_free_label()
                self.replace_label( cur_gap, new_value, gapy )
                cur_gap = new_value

    def swap_labels( self, lab, olab, frame ):
        """ Exchange two labels """
        self.tracking.swap_frame_id(lab, olab, frame)

    def swap_tracks( self, lab, olab, start_frame ):
        """ Exchange two tracks """
        ## split the two labels to unused value
        tmp_labels = self.get_free_labels( 2 )
        for i, laby in enumerate([lab, olab]):
            self.replace_label( laby, tmp_labels[i], start_frame )
            
        ## replace the two initial labels, in inversed order
        self.replace_label( tmp_labels[0], olab, start_frame )
        self.replace_label( tmp_labels[1], lab, start_frame )
        

    def update_changed_labels_img( self, img_before, img_after, added=True, removed=True ):
        """ Update tracks from changes between the two labelled images """
        if self.verbose > 1:
            print("Updating changed labels from images")
        indmodif = np.argwhere( img_before != img_after ).tolist()
        if len(indmodif) <= 0:
            return
        indmodif = tuple( np.array(indmodif).T )
        new_labels = img_after[indmodif]
        old_labels = img_before[indmodif]
        self.update_changed_labels( indmodif, new_labels, old_labels )

    
    def added_labels_oneframe( self, frame, img_before, img_after ):
        """ Update added tracks between the two labelled images at frame """
        ## Look for added labels
        added_labels = np.setdiff1d( img_after, img_before )
        self.tracking.add_one_frame( added_labels, frame, refresh=True )
    
    def removed_labels( self, img_before, img_after, frame=None ):
        """ Update removed tracks between the two labelled images """
        ## Look for added labels
        deleted_labels = np.setdiff1d( img_before, img_after )
        if frame is None:
            self.tracking.remove_tracks( deleted_labels )
        else:
            self.tracking.remove_one_frame( track_id=deleted_labels.tolist(), frame=frame, handle_gaps=self.forbid_gaps )

    def remove_label(self, label, force=False):
        """ Remove a given label if allowed """
        ut.changeLabel(self.seglayer, label, 0)
        self.tracking.remove_tracks(label)
        self.seglayer.refresh()

    def remove_labels(self, labels, force=False):
        """ Remove all allowed labels """
        inds = []
        for lab in labels:
            #if (force) or (not self.locked_label(label)):
            inds = inds + ut.getLabelIndexes(self.seglayer.data, lab, None)
        ut.setNewLabel(self.seglayer, inds, 0)
        self.tracking.remove_tracks(labels)
    
    def keep_labels(self, labels, force=True):
        """ Remove all other labels that are not in labels """
        inds = []
        toremove = list( set(self.tracking.get_track_list()) - set(labels) )
        #for lab in self.tracking.get_track_list():
        #    if lab not in labels:
                #if (force) or (not self.locked_label(label)):
        for lab in toremove:
            inds = inds + ut.getLabelIndexes(self.seglayer.data, lab, None)
        #        toremove.append(lab)
        ut.setNewLabel(self.seglayer, inds, 0)
        self.tracking.remove_tracks(toremove)

    def get_frame_features( self, frame ):
        """ Measure the label properties of given frame """
        return regionprops( self.seg[frame] )

    def updates_after_tracking( self ):
        """ When tracking has been done, update events, others """
        self.inspecting.get_divisions()

    #######################
    ## Classified cells options
    def get_all_groups(self, numeric=False):
        """ Add all groups info """
        if numeric:
            groups = [0]*self.nlabels()
        else:
            groups = ["None"]*self.nlabels()
        for igroup, gr in self.groups.keys():
            indexes = self.tracking.get_track_indexes(self.groups[gr])
            if numeric:
                groups[indexes] = igroup + 1
            else:
                groups[indexes] = gr
        return groups 
    
    def get_groups(self, labels, numeric=False):
        """ Add the group info of the given labels (repeated) """
        if numeric:
            groups = [0]*len(labels)
        else:
            groups = ["None"]*len(labels)
        for lab in np.unique(labels):
            gr = self.find_group( lab )
            if numeric:
                gr = self.groups.keys().index() + 1
            indexes = (np.argwhere(labels==lab)).flatten()
            for ind in indexes:
                groups[ind] = gr
        return groups 
    
    def cells_ingroup(self, labels, group):
        """ Put the cell "label" in group group, add it if new group """
        presents = self.has_labels( labels )
        labels = np.array(labels)[ presents ]
        if group not in self.groups.keys():
            self.groups[group] = []
            if self.outputing is not None:
                self.outputing.update_selection_list()
            self.editing.update_group_list( group )
        ## add only non present label(s)
        grlabels = self.groups[ group ]
        self.groups[ group ] = list( set( grlabels + labels.tolist()) )
        
    def find_group(self, label):
        """ Find in which group the label is """
        for gr, labs in self.groups.items():
            if label in labs:
                return gr
        return None

    def cell_removegroup(self, label):
        """ Detach the cell from its group """
        if not self.has_label(label):
            if self.verbose > 1:
                print("Cell "+str(label)+" missing")
        group = self.find_group(label)
        if group is not None:
            self.groups[group].remove(label)
            if len(self.groups[group]) <= 0:
                del self.groups[group]
                self.outputing.update_selection_list()

    def reset_groups(self):
        """ Remove all group information for all cells """
        self.groups = {} 
        self.outputing.update_selection_list()

    def draw_groups(self):
        """ Draw all the epicells colored by their group """
        grouped = np.zeros( self.seg.shape, np.uint8 ) 
        if (self.groups is None) or len(self.groups.keys()) == 0:
            return grouped
        for group, labels in self.groups.items():
            igroup = self.get_group_index(group) + 1
            np.place( grouped, np.isin( self.seg, labels ), igroup )
        return grouped

    def get_group_index(self, group):
        """ Get the index of group in the list of groups """
        igroup = (list(self.groups.keys())).index(group)
        return igroup
    
    ######### ROI
    def only_current_roi(self, frame):
        """ Put 0 everywhere outside the current ROI """
        roi_labels = self.editing.get_labels_inside()
        if roi_labels is None:
            return None
        # remove all other labels that are not in roi_labels
        roilab = np.copy(self.seg[frame])
        inds = []
        np.place(roilab, np.isin(roilab, roi_labels, invert=True), 0) 
        return roilab









"""
Tracking with laptrack package

Inspired from example https://github.com/yfukai/laptrack/blob/main/docs/examples/cell_segmentation.ipynb
"""

import napari
import numpy as np
import pandas as pd
import time

from laptrack import LapTrack
from laptrack import datasets
import epicure.Utils as ut
from napari.utils import progress

from napari.utils.notifications import show_info
#from multiprocessing.pool import ThreadPool as Pool
#from functools import partial
    
def squared_difference(a1, a2):
    """ Squared difference, normalized """
    return ((a1-a2)**2)/(max(a1, a2)**2)

def squared_distance(x1, y1, x2, y2):
    """ Return squared distance """
    return (x1-x2)**2 + (y1-y2)**2


class LaptrackCentroids():

    def __init__(self, track, epic):
        self.max_distance = 15
        self.splitting_cost = 1
        self.merging_cost = 1
        self.penal_area = 0
        self.penal_solidity = 0
        self.track = track
        self.epicure = epic
        self.inspecting = False
        self.suggesting = False
        self.region_properties = ["label", "frame", "centroid-0", "centroid-1", "area", "solidity"]

        self.lt = LapTrack( track_dist_metric=self.tracking_metric,
                track_cost_cutoff=1, splitting_dist_metric=self.tracking_nofeat_metric, splitting_cost_cutoff=self.splitting_cost, merging_dist_metric=self.tracking_nofeat_metric,
                merging_cost_cutoff=self.merging_cost, )
        
    def set_region_properties(self, with_extra=False):
        """ define the region properties used in the tracker """
        if with_extra:
            self.region_properties = ["label", "frame", "centroid-0", "centroid-1", "area", "solidity"]
        else:
            self.region_properties = ["label", "frame", "centroid-0", "centroid-1" ]

    def twoframes_track(self, df, labels):
        """ Do track between two frames, only track label """
        #start_time = time.time()
        track_df, split_df, merge_df = self.perform_track( df )
        #show_info("Performed in "+"{:.3f}".format((time.time()-start_time)/60)+" min")
        
        track_ids = [None]*len(labels)
        ## look for track id associated with label
        frame_df = track_df[track_df["frame"]==1]
        for i, row in frame_df.iterrows():
            tid = int(row["track_id"])
            curlabel = int(row["label"])
            ind = labels.index(curlabel)
            track_ids[ind] = tid

        tracklabels = [None]*len(labels)
        ## look for cell label associated with track_id in the first frame, if any
        frame_df = track_df[track_df["frame"]==0]
        frame_df = frame_df[np.isin(frame_df["track_id"].astype(int).values, track_ids)]
        for i, row in frame_df.iterrows():
            tid = int(row["track_id"])
            if (tid in track_ids):
                #xc = int(row["centroid-0"])
                #yc = int(row["centroid-1"])
                ind = track_ids.index(tid)
                label = int(row["label"])
                tracklabels[ind] = label #img[0, xc, yc]
        #show_info("Finished in "+"{:.3f}".format((time.time()-start_time)/60)+" min")
        return tracklabels

    def tracking_metric(self, c1, c2):
        """ Tracking cost metric: distance + feature penalties """
        dist = squared_distance(c1[2], c1[3], c2[2], c2[3])
        if dist > self.max_distance**2:
            return dist
        ## normalize all distances to combine them
        dist = dist / (self.max_distance**2)
        if self.penal_area > 0:
            dist += self.penal_area * squared_difference(c1[4], c2[4])
        if self.penal_solidity > 0:
            dist += self.penal_solidity * squared_difference(c1[5], c2[5])
        return dist
    
    def tracking_nofeat_metric(self, c1, c2):
        """ Tracking cost metric: distance, no penalties """
        dist = squared_distance(c1[2], c1[3], c2[2], c2[3])
        if dist > self.max_distance**2:
            return dist
        ## normalize all distances to combine them
        dist = dist / (self.max_distance**2)
        return dist

    def perform_track(self, regionprops_df):
        """ Do tracking with laptrack module """
        split_cost = False
        merg_cost = False
        #from laptrack import ParallelBackend
        #self.lt.parallel_backend = ParallelBackend.ray
        if self.splitting_cost > 0:
            split_cost = self.splitting_cost**2
        self.lt.splitting_cost_cutoff = split_cost
        if self.merging_cost > 0:
            merg_cost = self.merging_cost**2
        self.lt.merging_cost_cutoff = merg_cost
        track_df, split_df, merge_df = self.lt.predict_dataframe(
            regionprops_df,
            coordinate_cols=self.region_properties,
            only_coordinate_cols=False,
        )    
        track_df = track_df.reset_index()
        return track_df, split_df, merge_df

    def track_centroids(self, regionprops_df):
        """ Track cells based on their centroids positions + features penalties """
        ut.napari_info("Starting tracking with LapTrack centroids metrics...")
        return self.perform_track( regionprops_df )

    def inspect_oneframe(self, graph, trackdf):
        for track in np.unique(trackdf["track_id"]):
            tr = trackdf[trackdf["track_id"] == track]
            ## track is only on one frame, suspect
            if len(np.unique(tr["frame"])) == 1:
                # trackid + 1 as trackid starts as 0
                pos = (tr.iloc[0]["frame"], int(tr.iloc[0]["centroid-0"]), int(tr.iloc[0]["centroid-1"]))
                self.epicure.inspecting.add_event( pos, track+1, "tracking" )
                if self.track.suggesting:
                    if track in graph.keys():
                        sisters = []
                        refval = graph[track][0]
                        for key, val in graph.items():
                            if val[0] == refval:
                                sisters.append( key )
                        if len(sisters) == 2:
                            for sis in sisters:
                                self.epicure.add_suggestion( sis+1, refval+1 )





    ######## Cell trajectory
    def one_cell_movie(self, label):
        """ Isolate the movie of one cell """
        if label not in np.unique(self.seg):
            ut.show_error("Cell "+str(label)+" not found")
        bbox = ut.getBBoxLabel(self.seg, label)
        print(bbox)
        ## prepare the movie
        movshape = []
        extend = 1
        for i in range(3):
            if i > 0:
                extend = 1.1  ## keep a little outside the cell
            movshape.append( int((bbox[i+3]-bbox[i])*extend) )
        cellmovie = np.zeros( tuple(movshape) )
        ## get the center of the cell, copy sym and place at the center
        for frame in range(bbox[0], bbox[3]+1, 1):
            tseg = ut.get_frame()
    

    #### prev
    def reset_checked(self):
        """ erase all cells """
        self.epicells = {}
        self.groups = {}

    def is_checked(self, label):
        """ returns if label already checked or not """
        if self.epicells is None:
            return False
        return label in self.epicells.keys()

    def add_epicell(self, label, group="Checked"):
        """ Create a new (checked) branch, unmodifiable """
        epice = EpiCell(self)
        epice.create_epicell(label, group)
        self.epicells[label] = epice
        ## add new group if necessary
        if group not in self.groups:
            self.groups[group] = len(self.groups)+1
            self.outputing.update_selection_list()

    def remove_epicell(self, label):
        """ Uncheck the branch, remove it """
        if self.epicells.get(label) is None:
            print("No checked cell to remove")
        else:
            epice = self.epicells[label]
            self.epicells.pop(label) ## ? remove from dict 
            del epice
            print("Epicell removed")

    
    def reserve_labels(self, trackids, splitdf, mergedf):
        """ Unsure that trackids do not contain a checked label """
        if self.epicells is None:
            return trackids
        newval = np.max(trackids)+1
        for tid in np.unique(trackids):
            if tid in self.epicells.keys():
                #print("reserving "+str(tid))
                while newval in self.epicells.keys():
                    newval = newval + 1
                trackids[trackids==tid] = newval
                if len(splitdf) > 0:
                    splitdf = splitdf.replace(to_replace=tid, value=newval)
                if len(mergedf) > 0:
                    mergedf = mergedf.replace(to_replace=tid, value=newval)
                newval = newval + 1
        return trackids, splitdf, mergedf

    ############ Previous Get unused label
    def get_free_label(self):
        """ Get the smallest unused label """
        if self.unused_label is None:
            self.unused_label = ut.get_free_label(self.seglayer)
            return self.unused_label
        #if self.unused_label in np.unique(self.seglayer.data):
        #    self.update_free_label()
        return self.unused_label
    
    def free_label(self):
        """ Get the smallest unused label and update it """
        res = self.get_free_label()
        self.update_free_label()
        return res

    def relabel(self):
        """ Relabel all labels to have consecutive values """
        ut.relabel_layer(self.seglayer)
        self.unused_label = np.max(self.seglayer.data)+1

    def reset_free_label(self):
        """ Get the first smallest label """
        self.unused_label = ut.get_free_label(self.seglayer)

    def update_free_label(self):
        """ Update the value of smallest unused label """
        self.unused_label = self.unused_label + 1 #ut.get_next_label(self.seglayer, self.unused_label)

    def removed_label(self, label):
        """ Label has been removed, update what depended on it """
        if self.unused_label is None:
            self.unused_label = self.get_free_label()
        #else:
        #    if label < self.unused_label:
        #        self.unused_label = label

    ######### Replace a label by another one, handle all dependencies
    def replace_label(self, old_label, new_label, check_new_label=True):
        """ Replace label by new one, update everything """
        ## if needed, check if the new label is present and change it if yes
        if check_new_label:
            if new_label in np.unique(self.seglayer.data):
                self.replace_label(new_label, np.max(self.seglayer.data+1, False))
        ## change the label by the new one
        mask = self.seglayer.data == old_label
        np.place(self.seglayer.data, mask, new_label)


   ################"" Check
    def check_cell(self, event):
        """ Mark cell as checked and remove suspect """
        tframe = ut.current_frame(self.viewer, self.epicure.dim)
        #suglayer = self.viewer.layers["Suggestion"]
        #ut.setCellValue(checklayer, self.seglayer, event, 1, tframe, tframe)
        #ut.setCellValue(suglayer, self.seglayer, event, 0, tframe, tframe)
        ## Exonerate suspects in that cell
        self.epicure.suspecting.exonerate_from_event(event)
        ## Branch options
        label = ut.getCellValue( self.seglayer, event ) 
        group = self.check_group.text()
        self.epicure.add_epicell( label, group )
        if "Checked" in self.viewer.layers:
            self.epicure.draw_update_epicell( label, group )
    
    def uncheck_cell(self, event):
        """ Mark cell as uncheck """
        #checklayer = self.viewer.layers["CheckMap"]
        tframe = ut.current_frame(self.viewer, self.epicure.dim)
        #ut.setCellValue(checklayer, self.seglayer, event, 0, tframe, tframe)
        ## Branch options
        label = ut.getCellValue( self.seglayer, event ) 
        self.epicure.remove_epicell( label )
        if "Checked" in self.viewer.layers:
            self.epicure.draw_update_epicell( label, 0 )



import numpy as np
import os
import time
import math
from skimage.measure import regionprops, find_contours, regionprops_table
from skimage.segmentation import find_boundaries, expand_labels
from napari.utils.translations import trans
from napari.utils.notifications import show_info
from napari.utils import notifications as nt
from skimage.morphology import binary_dilation, disk
import pandas as pd
from epicure.laptrack_centroids import LaptrackCentroids

try:
    from skimage.graph import RAG
except:
    from skimage.future.graph import RAG  ## older version of scikit-image

def show_info(message):
    """ Display info in napari """
    nt.show_info(message)

def show_warning(message):
    """ Display a warning in napari (napari function show_warning doesn't work) """
    mynot = nt.Notification(message, nt.NotificationSeverity.WARNING)
    nt.notification_manager.dispatch(mynot)

def show_error(message):
    """ Display an error in napari (napari function show_error doesn't work) """
    mynot = nt.Notification(message, nt.NotificationSeverity.ERROR)
    nt.notification_manager.dispatch(mynot)

def show_debug(message):
    """ Display an info for debug in napari (napari function show_debug doesn't work) """
    print(message)

def show_documentation():
    import webbrowser
    webbrowser.open_new_tab("https://gitlab.pasteur.fr/gletort/epicure/-/wikis/Home")
    return

def show_documentation_page(page):
    import webbrowser
    webbrowser.open_new_tab("https://gitlab.pasteur.fr/gletort/epicure/-/wikis/"+page)
    return

def get_directory(imagepath):
    return os.path.dirname(imagepath)

def extract_names(imagepath, subname="epics", mkdir=True):
    imgname = os.path.splitext(os.path.basename(imagepath))[0]
    imgdir = os.path.dirname(imagepath)
    resdir = os.path.join(imgdir, subname)
    if (not os.path.exists(resdir)) and mkdir:
        os.makedirs(resdir)
    return imgname, imgdir, resdir

def extract_names_segmentation(segpath):
    """ Get the output directory and imagename from the segmentation filename """
    imgname = os.path.splitext(os.path.basename(segpath))[0]
    if imgname.endswith("_labels"):
        imgname = imgname[:(len(imgname)-7)]
    imgdir = os.path.dirname(segpath)
    return imgname, imgdir
    
def suggest_segfile(out, imgname):
    """ Check if a segmentation file from EpiCure already exists """
    segfile = os.path.join(out, imgname+"_labels.tif")
    if os.path.exists(segfile):
        return segfile
    else:
        return None
    
def get_filename(outdir, imgname):
    return os.path.join( outdir, imgname )

def napari_info(text):
    show_info(text)

def napari_shortcuts():
    """ Write main napari shortcuts list """
    text = "---- Main napari default shortcuts ----\n"
    text += " -- view options \n"
    text += "  <Ctrl+R> reset view \n"
    text += "  <Ctrl+Y> switch 2D/3D view mode \n"
    text += "  <Ctrl+G> switch Grid/Overlay view mode \n"
    text += "  <left arrow> got to previous frame \n"
    text += "  <right arrow> got to next frame \n"
    text += "\n"
    text += " -- labels options \n"
    text += "  <2> paint brush mode \n"
    text += "  <3> fill mode \n"
    text += "  <4> pick mode (select label) \n"
    text += "  <[> or <]> increase/decrease the paint brush size \n"
    text += "  <p> activate/deactivate preserve labels option \n"
    return text

def removeOverlayText(viewer):
    viewer.text_overlay.text = trans._("")
    viewer.text_overlay.visible = False

def getOverlayText(viewer):
    return viewer.text_overlay.text

def setOverlayText(viewer, text, size=12 ):
    viewer.text_overlay.text = trans._(text)
    viewer.text_overlay.visible = True
    viewer.text_overlay.font_size = size
    viewer.text_overlay.color = "white"
    viewer.text_overlay.opacity = 1

def showOverlayText(viewer, vis=None):
    if vis is None:
        viewer.text_overlay.visible = not viewer.text_overlay.visible
    else:
        viewer.text_overlay.visible = vis 

def reactive_bindings(layer, mouse_drag, key_map):
    """ Reactive the mouse and key event bindings on layer """
    layer.mouse_drag_callbacks = mouse_drag
    layer.keymap.update(key_map)

def clear_bindings(layer):
    """ Clear and returns the current event bindings on layer """
    old_mouse_drag = layer.mouse_drag_callbacks.copy()
    old_key_map = layer.keymap.copy()
    layer.mouse_drag_callbacks = []
    layer.keymap.clear()
    return old_mouse_drag, old_key_map

def is_binary( img ):
    """ Test if more than 2 values (skeleton or labelled image) """
    for frame in img:
        if len(np.unique(frame)) > 2:
            return False
    return True

def set_frame(viewer, frame):
    """ Set current frame """
    viewer.dims.set_point(0, frame)

def set_active_layer(viewer, layname):
    """ Set the current Napari active layer """
    if layname in viewer.layers:
        viewer.layers.selection.active = viewer.layers[layname]

def set_visibility(viewer, layname, vis):
    """ Set visibility of layer layname if exists """
    if layname in viewer.layers:
        viewer.layers[layname].visible = vis

def remove_layer(viewer, layname):
    if layname in viewer.layers:
        viewer.layers.remove(layname)

def remove_widget(viewer, widname):
    if widname in viewer.window._dock_widgets:
        wid = viewer.window._dock_widgets[widname]
        wid.setDisabled(True)
        try:
            wid.disconnect()
        except Exception:
            pass
        del viewer.window._dock_widgets[widname]
        wid.destroyOnClose()

def opentif(imagepath, verbose=True):
    import tifffile as tif
    img = tif.TiffFile(imagepath)
    metadata = img.imagej_metadata
    #print(metadata)
    scale = 1
    scalet = 1
    unitxy = "um"
    unitt = "min"
    nchan = -1
    if metadata is not None:
        if verbose:
            print(metadata)
        try:
            info = metadata["Info"]
            if info is not None: 
                metadatas = (info).splitlines()
                scale = float(metadatas[-4].split()[2])*1000000
        except:
            metadatas = None
        try:
            if metadata['physicalsizex'] is not None:
                scale = float(metadata['physicalsizex'])
            if metadata['finterval'] is not None:
                scalet = float(metadata['finterval'])
            if 'unit' in metadata:
                if metadata['unit'] is not None:
                    unitxy = metatdata['unit']
        except:
            metadatas = None
            #print(info)
        try:
            nchan = metadata["channels"]
            print("Nb chanels found: "+str(nchan))
        except:
            nchan = -1
    image = img.asarray()
    img.close()
    return image, nchan, scale, unitxy, scalet, unitt

def writeTif(img, imgname, scale, imtype, what=""):
    import tifffile
    if len(img.shape) == 2:
        tifffile.imwrite(imgname, np.array(img, dtype=imtype), imagej=True, resolution=[1./scale, 1./scale], metadata={'unit': 'um', 'axes': 'YX'})
    else:
        tifffile.imwrite(imgname, np.array(img, dtype=imtype), imagej=True, resolution=[1./scale, 1./scale], metadata={'unit': 'um', 'axes': 'TYX'}, compression="zstd")
    show_info(what+" saved in "+imgname)

def appendToTif(img, imgname):
    """ Append to RGB tif the current image """
    import tifffile
    tifffile.imwrite(imgname, img, photometric="rgb", append=True)

def getCellValue(label_layer, event):
    vis = label_layer.visible
    if vis == False:
        label_layer.visible = True
    label = label_layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
    label_layer.visible = vis
    return label

def setCellValue(layer, label_layer, event, newvalue, layer_frame=None, label_frame=None):
    # get concerned label (under the cursor), layer has to be visible for this
    vis = label_layer.visible
    if vis == False:
        label_layer.visible = True
    label = label_layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
    label_layer.visible = vis
    if label is not None and label > 0:
        # if the seg image is 2D (single frame), label_frame will be None
        if label_frame is not None and label_frame >= 0:
            ldata = label_layer.data[label_frame,:,:]
        else:
            ldata = label_layer.data
        # if the layer is 2D (single frame), layer_frame will be None
        if layer_frame is not None and layer_frame >= 0:
            #slice_coord = tuple(sc[keep_coords] for sc in slice_coord)
            cdata = layer.data[layer_frame,:,:]
        else:
            cdata = layer.data
            #slice_coord = tuple(sc[keep_coords] for sc in slice_coord)

        cdata[np.where(ldata==label)] = newvalue
        layer.refresh()
        return label
    
def get_skeleton( seg, verbose=0 ) :
    """ convert labels movie to skeleton (thin boundaries) """
    startt = start_time()
    skel = np.zeros(seg.shape, dtype="uint8")
    for z in range(seg.shape[0]):
        expz = expand_labels( seg[z], distance=1 )
        skel[z][(seg[z] == 0) *(expz > 0)] = 1
    if verbose > 0:
        show_duration(startt, header="Skeleton calculted in ")
    return skel


def setLabelValue(layer, label_layer, event, newvalue, layer_frame=None, label_frame=None):
    """ Change the label value under event position and returns its old value """
    ## get concerned label (under the cursor), layer has to be visible for this
    vis = label_layer.visible
    if vis == False:
        label_layer.visible = True
    label = label_layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
    label_layer.visible = vis
    
    if label > 0:
        inds = getLabelIndexes( label_layer.data, label, label_frame )
        setNewLabel(layer, inds, newvalue, add_frame=layer_frame)
        layer.refresh()
        return label
    return None

def getLabelIndexes(label_data, label, frame):
    """ Get the indixes at which label_layer is label for given frame """
    # if the seg image is 2D (single frame), frame will be None
    if frame is not None and frame >= 0:
        ldata = label_data[frame,:,:]
    else:
        ldata = label_data
    return np.argwhere( ldata==label ).tolist()

def getLabelIndexesInFrame(frame_data, label):
    """ Get the indexes at which frame data is label """
    # if the seg image is 2D (single frame), frame will be None
    return np.argwhere( frame_data==label ).tolist()

def changeLabel( label_layer, old_value, new_value ):
    """ replace the value of label old-value by new_value """
    index = np.argwhere( label_layer.data==old_value ).tolist()
    setNewLabel( label_layer, index, new_value )

def setNewLabel(label_layer, indices, newvalue, add_frame=None):
    """ Change the label of all the pixels indicated by indices """
    indexs = np.array(indices).T
    if add_frame is not None:
        indexs = np.vstack((np.repeat(add_frame, indexs.shape[1]), indexs))
    changed_indices = label_layer.data[tuple(indexs)] != newvalue
    inds = tuple(x[changed_indices] for x in indexs)
    oldvalues = label_layer.data[inds]
    if isinstance(newvalue, list):
        newvalue = np.array(newvalue)[np.where(changed_indices)[0]]
    label_layer.data_setitem( inds, newvalue )
    return inds, newvalue, oldvalues

def convert_coords( coord ):
    """ Get the time frame, and the 2D coordinates as int """
    int_coord = tuple(np.round(coord).astype(int))
    tframe = int(coord[0])
    int_coord = int_coord[1:3]
    return tframe, int_coord

def outerBBox2D(bbox, imshape, margin=0):
    if (bbox[0]-margin) <= 0:
        return True
    if (bbox[2]+margin) >= imshape[0]:
        return True
    if (bbox[1]-margin) <= 0:
        return True
    if (bbox[3]+margin) >= imshape[1]:
        return True
    return False

def isInsideBBox( bbox, obbox ):
    """ Check if bbox is included in obbox """
    if (bbox[0] >= obbox[0]) and (bbox[1] >= obbox[1]):
        return (bbox[2] <= obbox[2]) and (bbox[3] <= obbox[3])
    return False

def setBBox(position, extend, imshape):
    bbox = [int(position[0]-extend), int(position[1]-extend), int(position[2]-extend), int(position[0]+extend), int(position[1]+extend), int(position[2]+extend)]
    for i in range(3):
        if bbox[i] < 0:
            bbox[i] = 0
        if bbox[i+3] > imshape[i]:
            bbox[i+3] = imshape[i]
    return bbox

def setBBoxXY(position, extend, imshape):
    bbox = [int(position[0]), int(position[1]-extend), int(position[2]-extend), int(position[0]+1), int(position[1]+extend), int(position[2]+extend)]
    for i in range(3):
        if bbox[i] < 0:
            bbox[i] = 0
        if bbox[i+3] > imshape[i]:
            bbox[i+3] = imshape[i]
    return bbox

def getBBox2DFromPts(pts, extend, imshape):
    """ Get the bounding box surrounding all the points, plus a margin """
    arr = np.array(pts)
    ptsdim = arr.shape[1]
    if ptsdim == 2:
        bbox = [int(np.min(arr[:,0])), int(np.min(arr[:,1])), int(np.max(arr[:,0]))+1, int(np.max(arr[:,1]))+1]
    if ptsdim == 3:
        bbox = [int(np.min(arr[:,1])), int(np.min(arr[:,2])), int(np.max(arr[:,1]))+1, int(np.max(arr[:,2]))+1]
    if extend > 0:
        for i in range(2):
            bbox[i] = bbox[i] - extend
            if bbox[i] < 0:
                bbox[i] = 0
            bbox[i+2] = bbox[i+2] + extend
            if bbox[i+2] > imshape[i]:
                bbox[i+2] = imshape[i]
    return bbox

def getBBoxFromPts(pts, extend, imshape, outdim=None, frame=None):
    arr = np.array(pts)
    ## get if points are 2D or 3D
    ptsdim = arr.shape[1]
    ## if not imposed, output the same dimension as input points
    if outdim is None:
        outdim = ptsdim
    ## Get bounding box from points according to dimensions
    if ptsdim == 2:
        if outdim == 2:
            bbox = [int(np.min(arr[:,0])), int(np.min(arr[:,1])), int(np.max(arr[:,0]))+1, int(np.max(arr[:,1]))+1]
        else:
            bbox = [frame, int(np.min(arr[:,0])), int(np.min(arr[:,1])), frame+1, int(np.max(arr[:,0]))+1, int(np.max(arr[:,1]))+1]
    if ptsdim == 3:
        if outdim == 2:
            bbox = [int(np.min(arr[:,1])), int(np.min(arr[:,2])), int(np.max(arr[:,1]))+1, int(np.max(arr[:,2]))+1]
        else:
            bbox = [int(np.min(arr[:,0])), int(np.min(arr[:,1])), int(np.min(arr[:,2])), int(np.max(arr[:,0]))+1, int(np.max(arr[:,1]))+1, int(np.max(arr[:,2]))+1]
    if extend > 0:
        for i in range(outdim):
            bbox[(outdim==3)+i] = max( bbox[(outdim==3)+i] - extend, 0)
            bbox[(outdim==3)+i+outdim] = min(bbox[(outdim==3)+i+outdim] + extend, imshape[(outdim==3)+i] )
    return bbox

def inside_bounds( pt, imshape ):
    """ Chaque if given point is inside image limits """
    for i in range( len(pt) ):
        if (pt[i] < 0) or (pt[i] >= imshape[i]):
            return False
    return True

def extendBBox2D( bbox, extend_factor, imshape ):
    """ Extend bounding box with given margin """
    extend = max( bbox[2]-bbox[0], bbox[3]-bbox[1] )* extend_factor
    bbox = np.array(bbox)
    for i in range(2):
        bbox[i] = max( bbox[i] - extend, 0 )
        bbox[i+2] = min( bbox[i+2] + extend, imshape[i] )
    return bbox

def getBBox2D(img, label):
    """ Get bounding box of label """
    mask = (img==label)*1
    props = regionprops(mask)
    for prop in props:
        bbox = prop.bbox
        return bbox

def getPropLabel(img, label):
    """ Get the properties of label """
    props = regionprops(np.uint8(img==label))
    #print(props)
    return props[0]

def getBBoxLabel(img, label):
    """ Get bounding box of label """
    mask_ind = np.where(img==label)
    if len(mask_ind) <= 0:
        return None
    dim = len(img.shape)
    bbox = np.zeros(dim*2, int)
    for i in range(dim):
        bbox[i] = int(np.min(mask_ind[i]))
        bbox[i+dim] = int(np.max(mask_ind[i]))+1
    return bbox

def getBBox2DMerge(img, label, second_label): #, checkTouching=False):
    """ Get bounding box of two labels and check if they are in contact """
    mask = np.isin( img, [label, second_label] )
    props = regionprops(mask*1)
    return props[0].bbox, mask 

def checkTouchingLabels(img, label, second_label):
    """ Returns if labels are in contact (1-2 pixel away) """
    disk_one = disk(radius=1)
    maska = binary_dilation(img==label, footprint=disk_one)
    maskb = binary_dilation(img==second_label, footprint=disk_one)
    return np.any(maska & maskb)

def positionsIn2DBBox( positions, bbox ):
    """ Shift all the positions to their position inside the 2D bounding box """
    return [positionIn2DBBox( pos, bbox ) for pos in positions ]

def positions2DIn2DBBox( positions, bbox ):
    """ Shift all the positions to their position inside the 2D bounding box """
    return [position2DIn2DBBox( pos, bbox ) for pos in positions ]

def positionIn2DBBox(position, bbox):
    """ Returns the position shifted to its position inside the 2D bounding box """
    return (int(position[1]-bbox[0]), int(position[2]-bbox[1]))

def position2DIn2DBBox(position, bbox):
    """ Returns the position shifted to its position inside the 2D bounding box """
    return (int(position[0]-bbox[0]), int(position[1]-bbox[1]))

def toFullImagePos( indices, bbox ):
    return [ [ind[0]+bbox[0], ind[1]+bbox[1]] for ind in indices ]

def addFrameIndices( indices, frame ):
    return [ [frame, ind[0], ind[1]] for ind in indices ]

def shiftFrameIndices( indices, add_frame ):
    return [ [add_frame, ind[0], ind[1]] for ind in indices ]

def toFullMoviePos( indices, bbox, frame=None ):
    """ Replace indexes inside bounding box to full movie indexes """
    if frame is not None:
        indices = np.array(indices)
        frame_arr = np.repeat(frame, len(indices[:,0]))
        return np.stack([frame_arr, indices[:,0]+bbox[0], indices[:,1]+bbox[1]], axis=1)
    ## frame is None if indices contains already the frame position
    indices = np.array(indices)
    return np.stack([indices[:,0], indices[:,1]+bbox[0], indices[:,2]+bbox[1]], axis=1)

def cropBBox(img, bbox):
    if len(bbox) == 6:
        return img[bbox[0]:bbox[3], bbox[1]:bbox[4], bbox[2]:bbox[5]]
    return img[bbox[0]:bbox[2], bbox[1]:bbox[3]]


def crop_twoframes( img, bbox, frame ):
    """ Crop bounding box with two frames """
    return np.copy(img[(frame-1):(frame+1), bbox[0]:bbox[2], bbox[1]:bbox[3]])

def cropBBox2D(img, bbox):
    return img[bbox[0]:bbox[2], bbox[1]:bbox[3]]

def setValueInBBox2D(img, setimg, bbox):
    bbimg = img[bbox[0]:bbox[2], bbox[1]:bbox[3]] 
    bbimg[setimg>0]= setimg[setimg>0]

def addValueInBBox(img, addimg, bbox):
    img[bbox[0]:bbox[3], bbox[1]:bbox[4], bbox[2]:bbox[5]] = img[bbox[0]:bbox[3], bbox[1]:bbox[4], bbox[2]:bbox[5]] + addimg

def set_maxlabel(layer):
    layer.mode = "PAINT"
    layer.selected_label = np.max(layer.data)+1
    layer.refresh()

def set_label(layer, lab):
    layer.mode = "PAINT"
    layer.selected_label = lab
    layer.refresh()

def get_free_labels( used, nlab ):
    """ Get n-th unused label (not in used list) """
    maxlab = max(used)+1
    unused = list(set(range(1, maxlab)) - set(used))
    if nlab < len(unused):
        return unused[0:nlab]
    else:
        return unused+list(range(maxlab+1, maxlab+1+(nlab-len(unused))))

def get_next_label(layer, label):
    """ Get the next unused label starting from label """
    used = np.unique(layer.data)
    i = label+1
    while i < np.max(used):
        if i>0 and (i not in used):
            return i
        i = i + 1
    return i+1

def relabel_layer(layer):
    maxlab = np.max(layer.data)
    used = np.unique(layer.data)
    nlabs = len(used)
    if nlabs == maxlab:
        #print("already relabelled")
        return
    for j in range(1, nlabs+1):
        if j not in used:
            layer.data[layer.data==maxlab] = j
            maxlab = np.max(layer.data)
    show_info("Labels reordered")
    layer.refresh()

def inv_visibility(viewer, layername):
    """ Switch the visibility of a layer """
    if layername in viewer.layers:
        layer = viewer.layers[layername]
        layer.visible = not layer.visible

######## Measure labels

def summary_labels( seg ):
    """ Summary of labels (cells) measurements """
    props = regionprops(seg)
    avg_duration = 0
    avg_area = 0.0
    for prop in props:
        bbox = prop.bbox
        nz = 1
        if len(bbox)>4:
            nz = bbox[3]-bbox[0]
        avg_duration += nz
        avg_area += prop.area/nz
    return len(props), avg_duration/len(props), avg_area/len(props) 

def labels_in_cell( sega, segb, label ):
    """ Look at the labels of segb inside label from sega """
    cell = np.isin( sega, [label] )
    labelb = segb[ cell ]
    cell_area = np.sum( cell*1, axis=None )
    filled_area = np.sum( labelb>0 )
    nobj = len(np.unique( labelb ))
    if 0 in labelb:
        nobj = nobj - 1
    return nobj, (filled_area/cell_area), np.unique(labelb)


def match_labels( sega, segb ):
    """ Match the labels of the two segmentation images """
    region_properties = ["label", "centroid"]

    df0 = pd.DataFrame( regionprops_table( sega, properties=region_properties ) )
    df0["frame"] = 0
    df1 = pd.DataFrame( regionprops_table( segb, properties=region_properties ) )
    df1["frame"] = 1
    df = pd.concat([df0, df1])

    ## Link the two frames with LapTrack tracking
    laptrack = LaptrackCentroids(None, None)
    laptrack.max_distance = 10 
    laptrack.set_region_properties(with_extra=False)
    laptrack.splitting_cost = False ## disable splitting option
    laptrack.merging_cost = False ## disable merging option
    labels = list(np.unique(segb))
    if 0 in labels:
        labels.remove(0)
    parent_labels = laptrack.twoframes_track(df, labels)
    return parent_labels, labels

def labels_table( labimg, intensity_image=None, properties=None, extra_properties=None ):
    """ Returns the regionprops_table of the labels """
    if properties is None:
        properties = ['label', 'centroid']
    if intensity_image is not None:
        return regionprops_table( labimg, intensity_image=intensity_image, properties=properties, extra_properties=extra_properties )
    return regionprops_table( labimg, properties=properties, extra_properties=extra_properties )

def tuple_int(pos):
    if len(pos) == 3:
        return ( (int(pos[0]), int(pos[1]), int(pos[2])) )
    if len(pos) == 2:
        return ( (int(pos[0]), int(pos[1])) )

def prop_to_pos(prop, frame):
    return np.array( (frame, int(prop.centroid[0]), int(prop.centroid[1])) )

def current_frame(viewer):
    return int(viewer.cursor.position[0])

def distance( x, y ):
    """ 2d distance """
    return math.sqrt( (x[0]-y[0])*(x[0]-y[0]) + (x[1]-y[1])*(x[1]-y[1]) )

def interm_position( prop, a, b ):
    res = [0,0]
    res[0] = a[0] + prop*(b[0]-a[0])
    res[1] = a[1] + prop*(b[1]-a[1])
    return res

def nb_frames( seg, lab ):
    """ Return nb frames with label lab """
    labseg = seg==lab
    return np.sum( np.any(labseg, axis=(1,2)) )

def keep_orphans( img, comp_img, klabels ):
    """ Keep only labels that doesn't have a follower """
    valid_labels = np.setdiff1d(img[0], klabels)
    if (len(valid_labels)==1) and (valid_labels[0]==0):
        return
    labels = [val for val in valid_labels if (val!=0) and np.any(comp_img==val)]
    mask = np.isin(img, labels)
    img[mask] = 0

def keep_orphans_3d( img, klabels ):
    """ Keep only orphans labels or lab and olab """
    for label in np.unique(img[1]):
        if label not in klabels:
            if nb_frames( img, label ) == 2:
                img[img==label] = 0
    return img

def mean_nonzero( array ):
    nonzero = np.count_nonzero(array)
    if nonzero > 0:
        return np.sum(array)/nonzero
    return 0

def get_contours( binimg ):
    """ Return the contour of a binary shape """
    return find_contours( binimg )

###### Connectivity labels
def touching_labels( img, expand=3 ):
    """ Extends the labels to make them touch """
    return expand_labels( img, distance=expand )

def connectivity_graph( img, distance ):
    """ Returns the region adjancy graph of labels """
    touchlab = touching_labels( img, expand=distance )
    return RAG( touchlab, connectivity=2 )

####### Distance measures

def total_distance( pts_pos ):
    """ Total distance travelled by point with coordinates xpos and ypos """
    after = pts_pos[1:]
    before = pts_pos[:(-1)]
    disp = after - before
    return np.sum( np.sqrt( np.square(disp[:,0]) + np.square(disp[:,1]) ) )

def net_distance( pts_pos ):
    """ Net distance travelled by point with coordinates xpos and ypos """
    disp = pts_pos[len(pts_pos)-1] - pts_pos[0]
    return np.sum( np.sqrt( np.square(disp[0]) + np.square(disp[1]) ) )


###### Time measures
def start_time():
    return time.time()

def show_duration(start_time, header=None):
    if header is None:
        header = "Processed in "
    #show_info(header+"{:.3f}".format((time.time()-start_time)/60)+" min")
    print(header+"{:.3f}".format((time.time()-start_time)/60)+" min")

###### Preferences/shortcuts 

def shortcut_click_match( shortcut, event ):
    """ Test if the click event corresponds to the shortcut """
    button = 1
    if shortcut["button"] == "Right":
        button = 2
    if event.button != button:
        return False
    if "modifiers" in shortcut.keys():
        return set(list(event.modifiers)) == set(shortcut["modifiers"])
    else:
        if len(event.modifiers) > 0:
            return False
        return True
        
def print_shortcuts( shortcut_group ):
    """ Put to text the subset of shortcuts """
    text = ""
    for short_name, vals in shortcut_group.items():
        if vals["type"] == "key":
            text += "  <"+vals["key"]+"> "+vals["text"]+"\n"
        if vals["type"] == "click":
            modif = ""
            if "modifiers" in vals.keys():
                modifiers = vals["modifiers"]
                for mod in modifiers:
                    modif += mod+"-"
            text += "  <"+modif+vals["button"]+"-click> "+vals["text"]+"\n"
    return text






    ########## Suggestion functions 
    def init_suggestion(self):
        """ Initialize the layer that will contains propostion of tracks/segmentations """
        #suggestion = np.zeros(self.imgshape, dtype="uint16")
        #self.suggestion = self.viewer.add_labels(suggestion, blending="additive", name="Suggestion")
        self.suggestion = None
        
        @self.seglayer.mouse_drag_callbacks.append
        def click(layer, event):
            if event.type == "mouse_press":
                if 'Alt' in event.modifiers:
                    if event.button == 1:
                        pos = event.position
                        # alt+left click accept suggestion under the mouse pointer (in all frames)
                        self.accept_suggestion(pos)
    
    def locked_checked(self):
        return (("CheckMap" in self.viewer.layers) and (self.editing.lock_checked.isChecked()) )

    def reset_suggestion(self):
        """ erase all suggestions """
        self.suggestion.data = np.zeros(self.imgshape, dtype="uint16")

    def accept_suggestion(self, pos):
        """ Accept the modifications of the label at position pos (all the label) """
        #seglayer = self.viewer.layers["Segmentation"]
        label = self.suggestion.data[tuple(map(int, pos))]
        if label > 0:
            found = self.suggestion.data==label
            self.suspecting.exonerate( found, self.seglayer ) 
            indices = np.argwhere( found )
            ut.setNewLabel( self.seglayer, indices, label, add_frame=None )
            self.suggestion.data[self.suggestion.data==label] = 0
            self.suggestion.refresh()
    
    def add_suggestion(self, label, new_label):
        self.suggestion.data[self.seglayer.data == label] = new_label


        ### In tracking
        self.suggesting = False


        #self.check_suggestion = QCheckBox(text="Suggest corrections")
        #layout.addWidget(self.check_suggestion)
        #self.check_suggestion.stateChanged.connect( self.checked_suggestion )
        #self.checked_suggestion()
        #self.check_suggestion.setChecked(False)
    def checked_suggestion(self):
        self.suggesting = self.check_suggestion.isChecked()
       ut.set_visibility(self.viewer, "Suggestion", self.suggesting)



import numpy as np
from skimage import filters
from skimage.measure import regionprops, label
from skimage.morphology import binary_erosion, binary_dilation, disk
from qtpy.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QLabel, QComboBox
from napari.utils import progress
import epicure.Utils as ut
import epicure.epiwidgets as wid
import time

"""
    EpiCure - Inspection interface
    Handle supects, events detection layer
"""

class Inspecting(QWidget):
    
    def __init__(self, napari_viewer, epic):
        super().__init__()
        self.viewer = napari_viewer
        self.epicure = epic
        self.seglayer = self.viewer.layers["Segmentation"]
        self.border_cells = None    ## list of cells that are on the border (touch the background)
        self.eventlayer_name = "Events"
        self.events = None
        self.win_size = 10

        ## Print the current number of events
        self.nevents_print = QLabel("")
        self.update_nevents_display()
        
        self.create_eventlayer()
        layout = QVBoxLayout()
        layout.addWidget( self.nevents_print )
        
        show_line, self.show_divisions, self.show_suspects = wid.double_check( "Show divisions", True, None, "Show/hide the division events", "Show suspects", True, None, "Show/hide suspect events" )
        layout.addLayout( show_line )
        self.show_divisions.stateChanged.connect( self.show_hide_divisions )
        self.show_suspects.stateChanged.connect( self.show_hide_suspects )
        
        ## Handle division events
        #div_line, self.show_divisions = wid.button_check_line( btn="Update divisions", btn_func=self.get_divisions, check="Show divisions", checked=True, #checkfunc=None, descr_btn="Update the list of division events from the track graph", descr_check="Show/hide the division events", leftbtn=False )
        update_div_btn = wid.add_button( btn="Update divisions from graph", btn_func=self.get_divisions, descr="Update the list of division events from the track graph" )
        layout.addWidget(update_div_btn)
        #self.show_divisions.stateChanged.connect( self.show_hide_divisions )
        
        ### Reset: delete all events
        reset_color = self.epicure.get_resetbtn_color()
        reset_event_btn = wid.add_button( btn="Reset all events", btn_func=self.reset_all_events, descr="Delete all current events", color=reset_color )
        layout.addWidget(reset_event_btn)
        
        ## Error suggestions based on cell features
        outlier_line, self.outlier_vis, self.featOutliers = wid.checkgroup_help( "Outlier options", False, "Show/Hide outlier options panel", "event#frame-based-events", self.epicure.display_colors, "group" )
        layout.addLayout( outlier_line )
        self.create_outliersBlock() 
        layout.addWidget(self.featOutliers)
        
        ## Error suggestions based on tracks
        track_line, self.track_vis, self.eventTrack = wid.checkgroup_help( "Track options", True, "Show/hide track options", "event#track-based-events", self.epicure.display_colors, "group2" )
        self.create_tracksBlock() 
        layout.addLayout( track_line )
        layout.addWidget(self.eventTrack)
        
        ## Visualisation options
        disp_line, self.event_disp, self.displayevent = wid.checkgroup_help( "Display options", False, "Show/hide event display options panel", "event#visualisation", self.epicure.display_colors, "group3" )
        self.create_displayeventBlock() 
        layout.addLayout( disp_line )
        layout.addWidget(self.displayevent)
        
        self.setLayout(layout)
        self.key_binding()

    def key_binding(self):
        """ active key bindings for events options """
        sevents = self.epicure.shortcuts["Events"]
        self.epicure.overtext["events"] = "---- Events editing ---- \n"
        self.epicure.overtext["events"] += ut.print_shortcuts( sevents )
   
        @self.epicure.seglayer.mouse_drag_callbacks.append
        def handle_event(seglayer, event):
            if event.type == "mouse_press":
                ## remove a event
                if ut.shortcut_click_match( sevents["delete"], event ):
                    ind = ut.getCellValue( self.events, event ) 
                    if self.epicure.verbose > 1:
                        print("Removing clicked event, at index "+str(ind))
                    if ind is None:
                        ## click was not on a event
                        return
                    sid = self.events.properties["id"][ind]
                    if sid is not None:
                        self.exonerate_one(ind, remove_division=True)
                        self.update_nevents_display()
                    else:
                        if self.epicure.verbose > 1:
                            print("event with id "+str(sid)+" not found")
                    self.events.refresh()
                    return

                ## zoom on a event
                if ut.shortcut_click_match( sevents["zoom"], event ):
                    ind = ut.getCellValue( self.events, event ) 
                    if "id" not in self.events.properties.keys():
                        print("No event under click")
                        return
                    sid = self.events.properties["id"][ind]
                    if self.epicure.verbose > 1:
                        print("Zoom on event with id "+str(sid)+"")
                    self.zoom_on_event( event.position, sid )
                    return

        @self.epicure.seglayer.bind_key( sevents["next"]["key"], overwrite=True )
        def go_next(seglayer):
            """ Select next suspect event and zoom on it """
            num_event = int(self.event_num.value())
            nevents = self.nb_events()
            if num_event < 0:
                if self.event_sizenb_events( only_suspect=True ) == 0:
                    if self.epicure.verbose > 0:
                        print("No more suspect event")
                    return  
                else:
                    self.event_num.setValue(0)
            else:
                self.event_num.setValue( (num_event+1)%nevents )
            self.skip_nonsuspect_event( nevents, nevents )
            self.go_to_event()       

    def skip_nonsuspect_event( self, nevents, left ):
        """ Skip next event if not a suspect one (eg division) """
        index = int(self.event_num.value())
        if self.is_division( index ):
            index = (index + 1)%nevents
            self.event_num.setValue( index )
            self.skip_nonsuspect_event( nevents, left-1 )
        if left < 0:
            return

    def create_eventlayer(self):
        """ Create a point layer that contains the events """
        features = {}
        pts = []
        self.events = self.viewer.add_points( np.array(pts), properties=features, face_color="red", size = 10, symbol='x', name=self.eventlayer_name, )
        self.event_types = {}
        self.update_nevents_display()
        self.epicure.finish_update()

    def load_events(self, pts, features, event_types, symbols=None, colors=None):
        """ Load events data from file and reinitialize layer with it"""
        ut.remove_layer(self.viewer, self.eventlayer_name)
        if symbols is None:
            symbols = "x"
        if colors is None:
            colors = "red"
        symb = symbols
        self.events = self.viewer.add_points( np.array(pts), properties=features, face_color=colors, size = 10, symbol=symbols, name=self.eventlayer_name, )
        self.event_types = event_types
        self.update_nevents_display()
        self.epicure.finish_update()

        
    ############### Display event options
    def get_event_types( self ):
        """ Returns the list of possible event types """
        return list( self.event_types.keys() )

    def update_nevents_display( self ):
        """ Update the display of number of event"""
        text = str(self.nb_events(only_suspect=True))+" suspects \n" 
        text += str(self.nb_type("division"))+" divisions" 
        self.nevents_print.setText( text )

    def nb_events( self, only_suspect=False ):
        """ Returns current number of events """
        if self.events is None:
            return 0
        if self.events.properties is None:
            return 0
        if "score" not in self.events.properties:
            return 0
        if not only_suspect:
            return len(self.events.properties["score"])
        return ( len(self.events.properties["score"]) - self.nb_type("division") )

    def get_events_from_type( self, feature ):
        """ Return the list of events of a given type """
        if feature == "suspect":
            sub_features = self.suspect_subtypes()
            evts_id = []
            for feat in sub_features:
                evts_id.extend( eid for eid in self.event_types[ feat ] if eid not in evts_id )
            return list( evts_id )
        if feature in self.event_types:
            return self.event_types[ feature ]
        return []

    def nb_type( self, feature ):
        """ Return nb of event of given type """
        if self.events is None:
            return 0
        if (self.event_types is None) or (feature not in self.event_types):
            return 0
        return len(self.event_types[feature])

    def create_displayeventBlock(self):
        ''' Block interface of displaying event layer options '''
        disp_layout = QVBoxLayout()
        
        ## Color mode
        colorlay = QHBoxLayout()
        color_label = QLabel()
        color_label.setText("Color by:")
        colorlay.addWidget(color_label)
        self.color_choice = QComboBox()
        colorlay.addWidget(self.color_choice)
        self.color_choice.addItem("None")
        self.color_choice.addItem("score")
        self.color_choice.addItem("tracking-2->1")
        self.color_choice.addItem("tracking-1-2-*")
        self.color_choice.addItem("track-length")
        self.color_choice.addItem("division")
        self.color_choice.addItem("area")
        self.color_choice.addItem("solidity")
        self.color_choice.addItem("intensity")
        self.color_choice.addItem("tubeness")
        self.color_choice.currentIndexChanged.connect(self.color_events)
        disp_layout.addLayout(colorlay)
        
        sizelay, self.event_size = wid.slider_line( "Point size:", minval=0, maxval=50, step=1, value=10, slidefunc=self.display_event_size, descr="Choose the current point size display" ) 
        disp_layout.addLayout(sizelay)

        ### Interface to select a event and zoom on it
        chooselay, self.event_num = wid.ranged_value_line( label="event n°", minval=0, maxval=1000000, step=1, val=0, descr="Choose current event to display/remove" )
        disp_layout.addLayout(chooselay)
        go_event_btn = wid.add_button( "Go to event", self.go_to_event, "Zoom and display current event" )
        disp_layout.addWidget(go_event_btn)
        clear_event_btn = wid.add_button( "Remove current event", self.clear_event, "Delete current event from the list of events" )
        disp_layout.addWidget(clear_event_btn)
        
        ## all features
        self.displayevent.setLayout(disp_layout)
        self.displayevent.setVisible( self.event_disp.isChecked() )
       
    #####
    def reset_event_range(self):
        """ Reset the max num of event """
        nsus = len(self.events.data)-1
        if self.event_num.value() > nsus:
            self.event_num.setValue(0)
        self.event_num.setMaximum(nsus)

    def go_to_event(self):
        """ Zoom on the currently selected event """
        num_event = int(self.event_num.value())
        ## if reached the end of possible events
        if num_event >= self.nb_events():
            num_event = 0
            self.event_num.setValue(0)
        if num_event < 0:
            if self.nb_events() == 0:
                if self.epicure.verbose > 0:
                    print("No more event")
                return  
            else:
                self.event_num.setValue(0)
                num_event = 0      
        pos = self.events.data[num_event]
        event_id = self.events.properties["id"][num_event]
        self.zoom_on_event( pos, event_id )

    def get_event_infos( self, sid ):
        """ Get the properties of the event of given id """
        index = self.index_from_id( sid )
        pos = self.events.data[ index ]
        label = self.events.properties[ "label" ][index]
        return pos, label

    def zoom_on_event( self, event_pos, event_id ):
        """ Zoom on chose event at given position """
        self.viewer.camera.center = event_pos
        self.viewer.camera.zoom = 5
        self.viewer.dims.set_point( 0, int(event_pos[0]) )
        crimes = self.get_crimes(event_id)
        if self.epicure.verbose > 0:
            print("Suspected because of: "+str(crimes))

    def color_events(self):
        """ Color points by the selected mode """
        color_mode = self.color_choice.currentText()
        self.events.refresh_colors()
        if color_mode == "None":
            self.events.face_color = "white"
        elif color_mode == "score":
            self.set_colors_from_properties("score")
        else:
            self.set_colors_from_event_type(color_mode)
        self.events.refresh_colors()

    def suspect_subtypes( self ):
        """ Return the list of suspect-related event types """
        features = list( self.event_types.keys() )
        if "division" in features:
            features.remove( "division" )
        return features

    def show_subset_event( self, feature, show=True ):
        """ Show/hide a subset (type) of event """
        tmp_size = int(self.event_size.value())
        size = 0.1
        if show:
            size = tmp_size
        ## select the events of corresponding type
        self.events.selected_data = {}
        if feature == "suspect":
            ## take all possible features except non-suspect ones (division, extrusion..)
            features = self.suspect_subtypes()
        else:
            features = [feature]
        for feat in features:
            self.select_feature_event( feat )
        self.events.current_size = size
        ## reset selection and default size
        self.events.selected_data = {}
        self.events.current_size = tmp_size
        self.events.refresh()

    def select_feature_event( self, feature ):
        """ Add all event of given feature to currently selected data """
        if feature not in self.event_types:
            return
        posid = self.event_types[feature]
        for sid in posid:
            ind = self.index_from_id(sid)
            self.events.selected_data.add(ind)

    def set_colors_from_event_type(self, feature):
        """ Set colors from given event_type feature (eg area, tracking..) """
        if self.event_types.get(feature) is None:
            self.events.face_color="white"
            return
        posid = self.event_types[feature]
        colors = ["white"]*len(self.events.data)
        ## change the color of all the positive events for the chosen feature
        for sid in posid:
            ind = self.index_from_id(sid)
            if ind is not None:
                colors[ind] = (0.8,0.1,0.1)
        self.events.face_color = colors

    def set_colors_from_properties(self, feature):
        """ Set colors from given propertie (eg score, label) """
        ncols = (np.max(self.events.properties[feature]))
        color_cycle = []
        for i in range(ncols):
            color_cycle.append( (0.25+float(i/ncols*0.75), float(i/ncols*0.85), float(i/ncols*0.75)) )
        self.events.face_color_cycle = color_cycle
        self.events.face_color = feature
    
    def update_display(self):
        self.events.refresh()
        self.color_events()

    def get_current_settings(self):
        """ Returns current event widget parameters """
        disp = {}
        disp["Point size"] = int(self.event_size.value())
        disp["Outliers ON"] = self.outlier_vis.isChecked()
        disp["Track ON"] = self.track_vis.isChecked()
        disp["EventDisp ON"] = self.event_disp.isChecked()
        disp["Show divisions"] = self.show_divisions.isChecked()
        disp["Show suspects"] = self.show_suspects.isChecked()
        disp["Ignore border"] = self.ignore_borders.isChecked()
        disp["Flag length"] = self.check_length.isChecked()
        disp["length"] = self.min_length.text()
        disp["Check size"] = self.check_size.isChecked()
        disp["Check shape"] = self.check_shape.isChecked()
        disp["Get apparitions"] = self.get_apparition.isChecked()
        disp["Get disparitions"] = self.get_disparition.isChecked()
        disp["threshold disparition"] = self.threshold_disparition.text()
        disp["Min area"] = self.min_area.text()
        disp["Max area"] = self.max_area.text()
        disp["Current frame"] = self.feat_onframe.isChecked()
        return disp

    def apply_settings( self, settings ):
        """ Set the current state (display, widget) from preferences if any """
        for setting, val in settings.items():
            if setting == "Outliers ON":
                self.outlier_vis.setChecked( val ) 
            if setting == "Track ON":
                self.track_vis.setChecked( val ) 
            if setting =="EventDisp ON":
                self.event_disp.setChecked( val ) 
            if setting == "Point size":
                self.event_size.setValue( int(val) )
                #self.display_event_size()
            if setting == "Show divisions":
                self.show_divisions.setChecked( val )
                self.show_hide_divisions()
            if setting == "Show suspects":
                self.show_suspects.setChecked( val )        
                self.show_hide_suspects()
            if setting == "Ignore border":
                self.ignore_borders.setChecked( val )
            if setting == "Flag length":
                self.check_length.setChecked( val )
            if setting == "length":
                self.min_length.setText( val )
            if setting == "Check size":
                self.check_size.setChecked( val )
            if setting == "Check shape":
                self.check_shape.setChecked( val )
            if setting == "Get apparitions":
                self.get_apparition.setChecked( val )
            if setting == "Get disparitions":
                self.get_disparition.setChecked( val )
            if setting == "Threshold disparition":
                self.threshold_disparition.setText( val )
            if setting == "Min area":
                self.min_area.setText( val )
            if setting == "Max area":
                self.max_area.setText( val )
            if setting == "Current frame":
                self.feat_onframe.setChecked( val )
 

    def display_event_size(self):
        """ Change the size of the point display """
        size = int(self.event_size.value())
        self.events.size = size
        self.events.refresh()
        #### Depend on event type, to update

    ############### eventing functions
    def get_crimes(self, sid):
        """ For a given event, get its event_type(s) """
        crimes = []
        for feat in self.event_types.keys():
            if sid in self.event_types.get(feat):
                crimes.append(feat)
        return crimes

    def add_event_type(self, ind, sid, feature):
        """ Add 1 to the event_type score for given feature """
        #print(self.event_types)
        if self.event_types.get(feature) is None:
            self.event_types[feature] = []
        self.event_types[feature].append(sid)
        self.events.properties["score"][ind] = self.events.properties["score"][ind] + 1

    def first_event(self, pos, label, featurename):
        """ Addition of the first event (initialize all) """
        ut.remove_layer(self.viewer, "Events")
        features = {}
        sid = self.new_event_id()
        features["id"] = np.array([sid], dtype="uint16")
        features["label"] = np.array([label], dtype=self.epicure.dtype)
        features["score"] = np.array([0], dtype="uint8")
        pts = [pos]
        self.events = self.viewer.add_points( np.array(pts), properties=features, face_color="score", size = 10, symbol="x", name="Events", )
        self.add_event_type(0, sid, featurename)
        self.events.refresh()
        self.update_nevents_display()

    def add_event(self, pos, label, reason, symb="x", color="white", force=False, refresh=True):
        """ Add a event to the list, evented by a feature """
        if (not force) and (self.ignore_borders.isChecked()) and (self.border_cells is not None):
            tframe = int(pos[0])
            if label in self.border_cells[tframe]:
                return

        ## initialise if necessary
        if len(self.events.data) <= 0:
            self.first_event(pos, label, reason)
            return
        
        self.events.selected_data = []
       
       ## look if already evented, then add the charge
        num, sid = self.find_event(pos[0], label)
        if num is not None:
            ## event already in the list. For same crime ?
            if self.event_types.get(reason) is not None:
                if sid not in self.event_types[reason]:
                    self.add_event_type(num, sid, reason)
            else:
                self.add_event_type(num, sid, reason)
        else:
            ## new event, add to the Point layer
            ind = len(self.events.data)
            sid = self.new_event_id()
            self.events.add(pos)
            self.events.properties["label"][ind] = label
            self.events.properties["id"][ind] = sid
            self.events.properties["score"][ind] = 0
            self.add_event_type(ind, sid, reason)

        self.events.symbol.flags.writeable = True
        self.events.current_symbol = symb
        self.events.current_face_color = color
        if refresh:
            self.refresh_events()

    def refresh_events( self ):
        """ Refresh event view and text """
        self.events.refresh()
        self.update_nevents_display()
        self.reset_event_range()

    def new_event_id(self):
        """ Find the first unused id """
        sid = 0
        if self.events.properties.get("id") is None:
            return 0
        while sid in self.events.properties["id"]:
            sid = sid + 1
        return sid
    
    def reset_all_events(self):
        """ Remove all event_types """
        features = {}
        pts = []
        ut.remove_layer(self.viewer, "Events")
        self.events = self.viewer.add_points( np.array(pts), properties=features, face_color="red", size = 10, symbol='x', name="Events", )
        self.event_types = {}
        self.update_nevents_display()
        #self.update_nevents_display()

    def reset_event_type(self, feature, frame):
        """ Remove all event_types of given feature, for current frame or all if frame is None """
        if self.event_types.get(feature) is None:
            return
        idlist = self.event_types[feature].copy()
        for sid in idlist:
            ind = self.index_from_id(sid)
            if ind is not None:
                if frame is not None:
                    if int(self.events.data[ind][0]) == frame:
                        self.event_types[feature].remove(sid)
                        self.decrease_score(ind)
                else:
                    self.event_types[feature].remove(sid)
                    self.decrease_score(ind)
        self.events.refresh()
        self.update_nevents_display()

    def remove_event_types(self, sid):
        """ Remove all event_types of given event id """
        for listval in self.event_types.values():
            if sid in listval:
                listval.remove(sid)

    def decrease_score(self, ind):
        """ Decrease by one score of event at index ind. Delete it if reach 0"""
        self.events.properties["score"][ind] = self.events.properties["score"][ind] - 1
        if self.events.properties["score"][ind] == 0:
            self.exonerate_one( ind, remove_division=False )
            self.update_nevents_display()

    def index_from_id(self, sid):
        """ From event id, find the corresponding index in the properties array """
        for ind, cid in enumerate(self.events.properties["id"]):
            if cid == sid:
                return ind
        return None

    def id_from_index( self, ind ):
        """ From event index, returns it id """
        return self.events.properties["id"][ind]

    def find_event(self, frame, label):
        """ Find if there is already a event at given frame and label """
        events = self.events.data
        events_lab = self.events.properties["label"]
        for i, lab in enumerate(events_lab):
            if lab == label:
                if events[i][0] == frame:
                    return i, self.events.properties["id"][i]
        return None, None

    def init_suggestion(self):
        """ Initialize the layer that will contains propostion of tracks/segmentations """
        suggestion = np.zeros(self.seglayer.data.shape, dtype="uint16")
        self.suggestion = self.viewer.add_labels(suggestion, blending="additive", name="Suggestion")
        
        @self.seglayer.mouse_drag_callbacks.append
        def click(layer, event):
            if event.type == "mouse_press":
                if 'Alt' in event.modifiers:
                    if event.button == 1:
                        pos = event.position
                        # alt+left click accept suggestion under the mouse pointer (in all frames)
                        self.accept_suggestion(pos)
    
    def accept_suggestion(self, pos):
        """ Accept the modifications of the label at position pos (all the label) """
        seglayer = self.viewer.layers["Segmentation"]
        label = self.suggestion.data[tuple(map(int, pos))]
        found = self.suggestion.data==label
        self.exonerate( found, seglayer ) 
        indices = np.argwhere( found )
        ut.setNewLabel( seglayer, indices, label, add_frame=None )
        self.suggestion.data[self.suggestion.data==label] = 0
        self.suggestion.refresh()
        self.update_nevents_display()
    
    def exonerate_one(self, ind, remove_division=True):
        """ Remove one event at index ind """
        self.events.selected_data = [ind]
        sid = self.events.properties["id"][ind]
        if (remove_division) and (ind in self.event_types["division"]):
            self.epicure.tracking.remove_division( self.events.properties["label"][ind] )
        self.events.remove_selected()
        self.remove_event_types(sid)
        
    def clear_event(self):
        """ Remove the current event """
        num_event = int(self.event_num.value())
        self.exonerate_one( num_event, remove_division=True )
        self.update_nevents_display()

    def exonerate_from_event(self, event):
        """ Remove all events in the corresponding cell of position """
        label = ut.getCellValue( self.seglayer, event )
        if len(self.events.data) > 0:
            for ind, lab in enumerate(self.events.properties["label"]):
                if lab == label:
                    if self.events.data[ind][0] == event.position[0]:      
                        self.exonerate_one(ind, remove_division=True) 
        self.update_nevents_display()

    def exonerate(self, indices, seglayer):
        """ Remove events that have been corrected/cleared """
        seglabels = np.unique(seglayer.data[indices])
        selected = []
        if self.events.properties.get("label") is None:
            return
        for ind, lab in enumerate(self.events.properties["label"]):
            if lab in seglabels:
                ## label to remove from event list
                selected.append(ind)
        if len(selected) > 0:
            self.events.selected_data = selected
            self.events.remove_selected()
            self.update_nevents_display()
                

    #######################################"
    ## Outliers suggestion functions
    def show_outlierBlock(self):
        self.featOutliers.setVisible( self.outlier_vis.isChecked() )

    def create_outliersBlock(self):
        ''' Block interface of functions for error suggestions based on cell features '''
        feat_layout = QVBoxLayout()
        
        self.feat_onframe = wid.add_check( check="Only current frame", checked=True, check_func=None, descr="Search for outliers only in current frame" )
        feat_layout.addWidget(self.feat_onframe)
        
        ## area widget
        tarea_layout, self.min_area, self.max_area = wid.min_button_max( btn="< Area (pix^2) <", btn_func=self.event_area_threshold, min_val="0", max_val="2000", descr="Look for cell which size is outside the given area range" )
        feat_layout.addLayout( tarea_layout )
        
        ## solid widget
        feat_solid_line, self.fsolid_out = wid.button_parameter_line( btn="Solidity outliers", btn_func=self.event_solidity, value="3.0", descr_btn="Search for outliers in solidity value", descr_value="Inter-quartiles range factor to consider outlier" )
        feat_layout.addLayout( feat_solid_line )
        
        ## intensity widget
        feat_inten_line, self.fintensity_out = wid.button_parameter_line( btn="Intensity cytoplasm/junction", btn_func=self.event_intensity, value="1.0", descr_btn="Search for outliers in intensity ratio", descr_value="Ratio of intensity above which the cell looks suspect" )
        feat_layout.addLayout( feat_inten_line )
        
        ## tubeness widget
        feat_tub_line, self.ftub_out = wid.button_parameter_line( btn="Tubeness cytoplasm/junction", btn_func=self.event_tubeness, value="1.0", descr_btn="Search for outliers in tubeness ratio", descr_value="Ratio of tubeness above which the cell looks suspect" )
        feat_layout.addLayout( feat_tub_line )
        
        ## all features
        self.featOutliers.setLayout(feat_layout)
        self.featOutliers.setVisible( self.outlier_vis.isChecked() )
    
    def event_feature(self, featname, funcname ):
        """ event in one frame or all frames the given feature """
        onframe = self.feat_onframe.isChecked()
        if onframe:
            tframe = ut.current_frame(self.viewer)
            self.reset_event_type(featname, tframe)
            funcname(tframe)
        else:
            self.reset_event_type(featname, None)
            for frame in range(self.seglayer.data.shape[0]):
                funcname(frame)
        self.update_display()
        ut.set_active_layer( self.viewer, "Segmentation" )
    
    def inspect_outliers(self, tab, props, tuk, frame, feature):
        q1 = np.quantile(tab, 0.25)
        q3 = np.quantile(tab, 0.75)
        qtuk = tuk * (q3-q1)
        for sign in [1, -1]:
            #thresh = np.mean(tab) + sign * np.std(tab)*tuk
            if sign > 0:
                thresh = q3 + qtuk
            else:
                thresh = q1 - qtuk
            for i in np.where((tab-thresh)*sign>0)[0]:
                position = ut.prop_to_pos( props[i], frame )
                self.add_event( position, props[i].label, feature )
    
    def event_area_threshold(self):
        """ Look for cell's area below/above a threshold """
        self.event_feature( "area", self.event_area_threshold_oneframe )

    def event_area_threshold_oneframe( self, tframe ):
        """ Check if area is above/below given threshold """
        minarea = int(self.min_area.text())
        maxarea = int(self.max_area.text())
        frame_props = self.epicure.get_frame_features( tframe )
        for prop in frame_props:
            if (prop.area < minarea) or (prop.area > maxarea):
                position = ut.prop_to_pos( prop, tframe )
                self.add_event( position, prop.label, "area" )


    def event_area(self, state):
        """ Look for outliers in term of cell area """
        self.event_feature( "area", self.event_area_oneframe )
    
    def event_area_oneframe(self, frame):
        seglayer = self.seglayer.data[frame]
        props = regionprops(seglayer)
        ncell = len(props)
        areas = np.zeros((ncell,1), dtype="float")
        for i, prop in enumerate(props):
            if prop.label > 0:
                areas[i] = prop.area
        tuk = self.farea_out.value()
        self.inspect_outliers(areas, props, tuk, frame, "area")

    def event_solidity(self, state):
        """ Look for outliers in term ofz cell solidity """
        self.event_feature( "solidity", self.event_solidity_oneframe )

    def event_solidity_oneframe(self, frame):
        seglayer = self.seglayer.data[frame]
        props = regionprops(seglayer)
        ncell = len(props)
        sols = np.zeros((ncell,1), dtype="float")
        for i, prop in enumerate(props):
            if prop.label > 0:
                sols[i] = prop.solidity
        tuk = float(self.fsolid_out.text())
        self.inspect_outliers(sols, props, tuk, frame, "solidity")
    
    def event_intensity(self, state):
        """ Look for abnormal intensity inside/periph ratio """
        self.event_feature( "intensity", self.event_intensity_oneframe )
    
    def event_intensity_oneframe(self, frame):
        seglayer = self.seglayer.data[frame]
        intlayer = self.viewer.layers["Movie"].data[frame] 
        props = regionprops(seglayer)
        for i, prop in enumerate(props):
            if prop.label > 0:
                self.test_intensity( intlayer, prop, frame )
    
    def test_intensity(self, inten, prop, frame):
        """ Test if intensity inside is much smaller than at periphery """
        bbox = prop.bbox
        intbb = inten[bbox[0]:bbox[2], bbox[1]:bbox[3]]
        footprint = disk(radius=self.epicure.thickness)
        inside = binary_erosion(prop.image, footprint)
        ininten = np.mean(intbb*inside)
        dil_img = binary_dilation(prop.image, footprint)
        periph = dil_img^inside
        periphint = np.mean(intbb*periph)
        if (periphint<=0) or (ininten/periphint > float(self.fintensity_out.text())):
            position = ( frame, int(prop.centroid[0]), int(prop.centroid[1]) )
            self.add_event( position, prop.label, "intensity" )
    
    def event_tubeness(self, state):
        """ Look for abnormal tubeness inside vs periph """
        self.event_feature( "tubeness", self.event_tubeness_oneframe )
    
    def event_tubeness_oneframe(self, frame):
        seglayer = self.seglayer.data[frame]
        mov = self.viewer.layers["Movie"].data[frame]
        sated = np.copy(mov)
        sated = filters.sato(sated, black_ridges=False)
        props = regionprops(seglayer)
        for i, prop in enumerate(props):
            if prop.label > 0:
                self.test_tubeness( sated, prop, frame )

    def test_tubeness(self, sated, prop, frame):
        """ Test if tubeness inside is much smaller than tubeness on periph """
        bbox = prop.bbox
        satbb = sated[bbox[0]:bbox[2], bbox[1]:bbox[3]]
        footprint = disk(radius=self.epicure.thickness)
        inside = binary_erosion(prop.image, footprint)
        intub = np.mean(satbb*inside)
        periph = prop.image^inside
        periphtub = np.mean(satbb*periph)
        if periphtub <= 0:
            position = ( frame, int(prop.centroid[0]), int(prop.centroid[1]) )
            self.add_event( position, prop.label, "tubeness" )
        else:
            if intub/periphtub > float(self.ftub_out.text()):
                position = ( frame, int(prop.centroid[0]), int(prop.centroid[1]) )
                self.add_event( position, prop.label, "tubeness" )


############# event based on track

    def show_tracksBlock(self):
        self.eventTrack.setVisible( self.track_vis.isChecked() )

    def create_tracksBlock(self):
        ''' Block interface of functions for error suggestions based on tracks '''
        track_layout = QVBoxLayout()
        
        self.ignore_borders = wid.add_check( "Ignore cells on border", False, None, "When adding suspect, don't add it if the cell is touching the border" )
        track_layout.addWidget(self.ignore_borders)
        
        ## Look for sudden appearance of tracks
        self.get_apparition = wid.add_check( "Flag track apparition", True, None, "Add a suspect if a track appears in the middle of the movie (not on border)" )
        track_layout.addWidget(self.get_apparition)
       
        ## Look for sudden disappearance of tracks
        disp_line, self.get_disparition, self.threshold_disparition = wid.check_value( check="Flag track disparition", checkfunc=None, checked=True, value="50", descr="Add a suspect if a track disappears (not last frame, not border) and the cell area is above threshold", label="if cell area above" )
        track_layout.addLayout( disp_line )

        ## track length event_types
        ilengthlay, self.check_length, self.min_length = wid.check_value( check="Flag tracks smaller than", checkfunc=None, checked=True, value="1", descr="Add a suspect event for each track smaller than chosen value (in number of frames)" )
        track_layout.addLayout(ilengthlay)
        
        ## Variability in feature event_type
        sizevar_line, self.check_size, self.size_variability = wid.check_value( check="Size variation", checkfunc=None, checked=False, value="1", descr="Add a suspect if the size of the cell varies suddenly in the track" )
        track_layout.addLayout( sizevar_line )
        shapevar_line, self.check_shape, self.shape_variability = wid.check_value( check="Shape variation", checkfunc=None, checked=False, value="2.0", descr="Add a suspect if the shape of the cell varies suddenly in the track" )
        track_layout.addLayout( shapevar_line )

        ## merge/split combinaisons 
        track_btn = wid.add_button( btn="Inspect track", btn_func=self.inspect_tracks, descr="Start track analysis to look for suspects based on selected features" )
        track_layout.addWidget(track_btn)
        
        ## all features
        self.eventTrack.setLayout(track_layout)
        self.eventTrack.setVisible( self.track_vis.isChecked() )

    def reset_tracking_event(self):
        """ Remove events from tracking """
        self.reset_event_type("tracking-1-2-*", None)
        self.reset_event_type("tracking-2->1", None)
        self.reset_event_type("track-length", None)
        self.reset_event_type("track-size", None)
        self.reset_event_type("track-shape", None)
        self.reset_event_type("track-apparition", None)
        self.reset_event_type("track-disparition", None)
        self.reset_event_range()

    def track_length(self):
        """ Find all cells that are only in one frame """
        max_len = int(self.min_length.text())
        labels, lengths, positions = self.epicure.tracking.get_small_tracks( max_len )
        for label, nframe, pos in zip(labels, lengths, positions):
            if self.epicure.verbose > 2:
                print("event track length "+str(nframe)+": "+str(label)+" frame "+str(pos[0]) )
            self.add_event(pos, label, "track-length", refresh=False)
        self.refresh_events()

    def inspect_tracks(self):
        """ Look for suspicious tracks """
        self.viewer.window._status_bar._toggle_activity_dock(True)
        ut.set_visibility( self.viewer, "Events", True )
        progress_bar = progress( total=8 )
        progress_bar.update(0)
        self.reset_tracking_event()
        progress_bar.update(1)
        if self.ignore_borders.isChecked():
            progress_bar.set_description("Identifying border cells")
            self.get_border_cells()
        progress_bar.update(2)
        tracks = self.epicure.tracking.get_track_list()
        if self.check_length.isChecked():
            progress_bar.set_description("Identifying too small tracks")
            self.track_length()
        progress_bar.update(3)
        progress_bar.set_description("Inspect tracks 2->1")
        self.track_21()
        progress_bar.update(4)
        if (self.check_size.isChecked()) or self.check_shape.isChecked():
            progress_bar.set_description("Inspect track features")
            self.track_features()
        progress_bar.update(5)
        if self.get_apparition.isChecked():
            progress_bar.set_description("Check new track apparition")
            self.track_apparition( tracks )
        progress_bar.update(6)
        if self.get_disparition.isChecked():
            progress_bar.set_description("Check track disparition")
            self.track_disparition( tracks, progress_bar )
        progress_bar.update(7)
        progress_bar.close()
        self.viewer.window._status_bar._toggle_activity_dock(False)
        ut.set_active_layer( self.viewer, "Segmentation" )

    def track_apparition( self, tracks ):
        """ Check if some track appears suddenly (in the middle of the movie and not by division) """
        start_time = time.time()
        ## remove track on first frame
        ctracks = list( set(tracks) - set( self.epicure.tracking.get_tracks_on_frame( 0 ) ) )
        graph = self.epicure.tracking.graph
        for i, track_id in enumerate( ctracks) :
            fframe = self.epicure.tracking.get_first_frame( track_id )
            ## If on the border, ignore
            outside = self.epicure.cell_on_border( track_id, fframe )
            if outside:
                continue
            ## Not on border, check if potential division
            if (graph is not None) and (track_id in graph.keys()):
                continue
            ## event apparition
            posxy = self.epicure.tracking.get_position( track_id, fframe )
            if posxy is not None:
                pos = [ fframe, posxy[0], posxy[1] ]
                if self.epicure.verbose > 2:
                    print("Appearing track: "+str(track_id)+" at frame "+str(fframe) )
                self.add_event(pos, track_id, "tracking-apparition", refresh=False)
        self.refresh_events()
        if self.epicure.verbose > 1:
            ut.show_duration( start_time, "Tracks apparition took " )
                
    def track_disparition( self, tracks, progress_bar ):
        """ Check if some track disappears suddenly (in the middle of the movie and not by division) """
        start_time = time.time()
        ## Track disappears in the movie, not last frame
        ctracks = list( set(tracks) - set( self.epicure.tracking.get_tracks_on_frame( self.epicure.nframes-1 ) ) )
        graph = self.epicure.tracking.graph
        threshold_area = float(self.threshold_disparition.text())
        sub_bar = progress( total = len( ctracks ), desc="Check non last frame tracks", nest_under = progress_bar )
        for i, track_id in enumerate( ctracks ):
            sub_bar.update( i )
            lframe = self.epicure.tracking.get_last_frame( track_id )
            ## If on the border, ignore
            outside = self.epicure.cell_on_border( track_id, lframe )
            if outside:
                continue
            ## Not on border, check if potential division
            if self.epicure.tracking.is_parent( track_id ):
                continue
            ## check if the cell area is below the threshold, then considered as ok (likely extrusion)
            if (threshold_area > 0):
                cell_area = self.epicure.cell_area( track_id, lframe )
                if cell_area < threshold_area:
                    continue
            ## event disparition
            posxy = self.epicure.tracking.get_position( track_id, lframe )
            if posxy is not None:
                pos = [ lframe, posxy[0], posxy[1] ]
                if self.epicure.verbose > 2:
                    print("Disappearing track: "+str(track_id)+" at frame "+str(lframe) )
                self.add_event(pos, track_id, "tracking-disparition", refresh=False)
        sub_bar.close()
        self.refresh_events()
        if self.epicure.verbose > 1:
            ut.show_duration( start_time, "Tracks disparition took " )

    def track_21(self):
        """ Look for event track: 2->1 """
        if self.epicure.tracking.tracklayer is None:
            ut.show_error("No tracking done yet!")
            return

        graph = self.epicure.tracking.graph
        if graph is not None:
            for child, parent in graph.items():
                ## 2->1, merge, event
                if len(parent) == 2:
                    onetwoone = False
                    ## was it only one before ?
                    if (parent[0] in graph.keys()) and (parent[1] in graph.keys()):
                        if graph[parent[0]][0] == graph[parent[1]][0]:
                            pos = self.epicure.tracking.get_mean_position([parent[0], parent[1]])
                            if pos is not None:
                                if self.epicure.verbose > 1:
                                    print("event 1->2->1 track: "+str(graph[parent[0]][0])+"-"+str(parent)+"-"+str(child)+" frame "+str(pos[0]) )
                                self.add_event(pos, parent[0], "tracking-1-2-*")
                                onetwoone = True
                
                    if not onetwoone:
                        pos = self.epicure.tracking.get_mean_position(child, only_first=True)     
                        if pos is not None:
                            if self.epicure.verbose > 2:
                                print("event 2->1 track: "+str(parent)+"-"+str(child)+" frame "+str(int(pos[0])) )
                            self.add_event(pos, parent[0], "tracking-2->1", refresh=False)
                        else:
                            if self.epicure.verbose > 1:
                                print("Something weird, "+str(child)+" mean position")

        self.epicure.finish_update()
        self.refresh_events()

    def get_border_cells(self):
        """ Return list of cells that are at the border (touching background) """
        self.border_cells = dict()
        for tframe in range(self.epicure.nframes):
            img = self.epicure.seg[tframe]
            self.border_cells[tframe] = self.get_border_cells_frame(img)      
        
    def get_border_cells_frame(self, imframe):
        """ Return cells on border in current image """ 
        height = self.epicure.imgshape2D[1]
        width = self.epicure.imgshape2D[0]
        labels = list( np.unique( imframe[ :, 0:2 ] ) )   ## top border
        labels += list( np.unique( imframe[ :, (height-2): ] ) )   ## bottom border
        labels += list( np.unique( imframe[ 0:2,] ) )   ## left border
        labels += list( np.unique( imframe[ (width-2):,] ) )   ## right border
        #graph = ut.connectivity_graph( imframe, distance=3 )
        #adj_bg = []
        #if 0 in graph.nodes:
        #    adj_bg = list( graph.adj[0 ])
        #return adj_bg
        return labels

    def get_divisions( self ):
        """ Get and add divisions from the tracking graph """
        self.reset_event_type( "division", frame=None )
        graph = self.epicure.tracking.graph
        divisions = {}
        ## Go through the graph and fill all division by parents
        if graph is not None:
            for child, parent in graph.items():
                ## 1 parent, potential division
                if (isinstance(parent, int)) or (len(parent) == 1):
                    if isinstance( parent, list ):
                        par = parent[0]
                    else:
                        par = parent
                    if par not in divisions:
                        divisions[par] = [child]
                    else:
                        divisions[par].append(child)

        ## Add all the divisions in the event list
        for parent, childs in divisions.items():
            indexes = self.epicure.tracking.get_track_indexes(childs)
            if len(indexes) <= 0:
                ## something wrong in the graph or in the tracks, ignore for now
                continue
            ## get the average first position of the childs just after division
            pos = self.epicure.tracking.mean_position(indexes, only_first=True)     
            self.add_event(pos, parent, "division", symb="o", color="#0055ffff", force=True)
        ## Update display to show/hide the divisions
        self.show_hide_divisions()
        self.epicure.finish_update()

    def show_hide_divisions( self ):
        """ Show or hide division events """
        self.show_subset_event( "division", self.show_divisions.isChecked() )

    def show_hide_suspects( self ):
        """ Show or hide suspect events """
        self.show_subset_event( "suspect", self.show_suspects.isChecked() )

    def add_division( self, labela, labelb, parent, frame ):
        """ Add a division event given the two daughter labels, the parent one and frame of division """
        indexes = self.epicure.tracking.get_index( [labela, labelb], frame )
        indexes = indexes.flatten()
        pos = self.epicure.tracking.mean_position( indexes )
        self.events.selected_data = {}
        if self.show_divisions.isChecked():
            self.events.current_size = int(self.event_size.value())
        else:
            self.events.current_size = 0.1
        self.add_event( pos, parent, "division", symb="o", color="#0055ffff", force=True )
        self.events.selected_data = {}
        self.events.current_size = int(self.event_size.value())
        ## check if there are suspect events to remove, cleared by the division
        if parent is not None:
            ## check eventual parent event
            num, sid = self.find_event(  pos[0]-1, parent )
            if num is not None:
                if self.is_end_event( sid ):
                    ## the parent event correspond to a potential end of track, remove it
                    ind = self.index_from_id( sid )
                    self.exonerate_one( ind, remove_division=False )
                    if self.epicure.verbose > 0:
                        print( "Removed suspect event of parent cell "+str(parent)+" cleared by the division flag" )
            ## check each child suspect if cleared by the new division 
            for child in [labela, labelb]:
                num, sid = self.find_event( pos[0], child )
                if num is not None:
                    if self.is_begin_event( sid ):
                        ## the child event correspond to a potential begin of track, remove it
                        ind = self.index_from_id( sid )
                        self.exonerate_one( ind, remove_division=False )
                        if self.epicure.verbose > 0:
                            print( "Removed suspect event of daughter cell "+str(child)+" cleared by the division flag" )
            self.update_nevents_display()


    def is_division( self, ind ):
        """ Return if the event of current index is a division """
        return ("division" in self.event_types) and (self.id_from_index(ind) in self.event_types["division"])

    def is_begin_event( self, sid ):
        """ Return True if the event has a type corresponding to begin of a track (too small or appearing) """
        beg_events = ["tracking-apparition", "track-length"]
        for event in beg_events:
            if event in self.event_types:
                if sid in self.event_types[event]:
                    return True
        return False

    def is_end_event( self, sid ):
        """ Return True if the event has a type corresponding to end of a track (too small or disappearing) """
        end_events = ["tracking-disparition", "track-length"]
        for event in end_events:
            if event in self.event_types:
                if sid in self.event_types[event]:
                    return True
        return False
        
    def track_features(self):
        """ Look at outliers in track features """
        track_ids = self.epicure.tracking.get_track_list()
        features = []
        featType = {}
        if self.check_size.isChecked():
            features = features + ["Area", "Perimeter"]
            featType["Area"] = "size"
            featType["Perimeter"] = "size"
            size_factor = float(self.size_variability.text())
        if self.check_shape.isChecked():
            features = features + ["Eccentricity", "Solidity"]
            featType["Eccentricity"] = "shape"
            featType["Solidity"] = "shape"
            shape_factor = float(self.shape_variability.text())
        for tid in track_ids:
            track_indexes = self.epicure.tracking.get_track_indexes( tid )
            ## track should be long enough to make sense to look for outlier
            if len(track_indexes) > 3:
                track_feats = self.epicure.tracking.measure_features( tid, features )
                for feature, values in track_feats.items():
                    if featType[feature] == "size":
                        factor = size_factor
                    if featType[feature] == "shape":
                        factor = shape_factor
                    outliers = self.find_jump( values, factor=factor )
                    for out in outliers:
                        tdata = self.epicure.tracking.get_frame_data( tid, out )
                        if self.epicure.verbose > 1:
                            print("event track "+feature+": "+str(tdata[0])+" "+" frame "+str(tdata[1]) )
                        self.add_event(tdata[1:4], tid, "track_"+featType[feature])

    def find_jump( self, tab, factor=1 ):
        """ Detect brutal jump in the values """
        jumps = []
        tab = np.array(tab)
        diff = tab[:(len(tab)-2)] - 2*tab[1:(len(tab)-1)] + tab[2:]
        diff = [(tab[1]-tab[0])] + diff.tolist() + [tab[len(tab)-1]-tab[len(tab)-2]] 
        avg = (tab[:(len(tab)-2)] + tab[2:])/2
        avg = [(tab[1]+tab[0])/2] + avg.tolist() + [(tab[len(tab)-1]+tab[len(tab)-2])/2]
        eps = 0.000000001
        diff = np.array(diff, dtype=np.float32)
        avg = np.array(avg, dtype=np.float32)
        diff = abs(diff+eps)/(avg+eps)
        ## keep only local max above threshold
        for i, diffy in enumerate(diff):
            if (i>0) and (i<len(diff)-1):
                if diffy > factor:
                    if (diffy > diff[i-1]) and (diffy > diff[i+1]):
                        jumps.append(i)
            else:
                if diffy > factor:
                    jumps.append(i)
        #jumps = (np.where( diff > factor )[0]).tolist()
        return jumps

    def find_outliers_tuk( self, tab, factor=3, below=True, above=True ):
        """ Returns index of outliers from Tukey's like test """
        q1 = np.quantile(tab, 0.2)
        q3 = np.quantile(tab, 0.8)
        qtuk = factor * (q3-q1)
        outliers = []
        if below:
            outliers = outliers + (np.where((tab-q1+qtuk)<0)[0]).tolist()
        if above:
            outliers = outliers + (np.where((tab-q3-qtuk)>0)[0]).tolist()
        return outliers

    def weirdo_area(self):
        """ look at area trajectory for outliers """
        track_df = self.epicure.tracking.track_df
        for tid in np.unique(track_df["track_id"]):
            rows = track_df[track_df["track_id"]==tid].copy()
            if len(rows) >= 3:
                rows["smooth"] = rows.area.rolling(self.win_size, min_periods=1).mean()
                rows["diff"] = (rows["area"] - rows["smooth"]).abs()
                rows["diff"] = rows["diff"].div(rows["smooth"])
                if self.epicure.verbose > 2:
                    print(rows)





"""
Tracking with laptrack package

Inspired from example https://github.com/yfukai/laptrack/blob/main/docs/examples/cell_segmentation.ipynb
"""

import napari
import numpy as np
import pandas as pd
from skimage.measure import regionprops_table
import time

from laptrack import OverLapTrack
from laptrack import datasets
import epicure.Utils as ut
from napari.utils import progress

from napari.utils.notifications import show_info
#from multiprocessing.pool import ThreadPool as Pool
#from functools import partial
    

class LaptrackOverlaps():

    def __init__(self, track, epic):
        self.splitting_cost = 1
        self.cost_cutoff = 0.9
        self.merging_cost = 1
        self.track = track
        self.epicure = epic
        self.inspecting = False
        self.suggesting = False
        

    def twoframes_track(self, img_labels, labels):
        """ Do track between two frames, only track label """
        #start_time = time.time()
        track_df, split_df, merge_df = self.perform_track( img_labels )
        #show_info("Performed in "+"{:.3f}".format((time.time()-start_time)/60)+" min")
        
        track_ids = [None]*len(labels)
        ## look for track id associated with label
        for i, row in track_df.iterrows():
            frame = int(row["frame"])
            if frame == 1:
                tid = int(row["track_id"])
                curlabel = int(row["label"])
                ind = labels.index(curlabel)
                track_ids[ind] = tid

        tracklabels = [None]*len(labels)
        ## look for cell label associated with track_id in the first frame, if any
        for i, row in track_df.iterrows():
            frame = int(row["frame"])
            tid = int(row["track_id"])
            if (tid in track_ids) and (frame==0):
                ind = track_ids.index(tid)
                label = int(row["label"])
                tracklabels[ind] = label 
        #show_info("Finished in "+"{:.3f}".format((time.time()-start_time)/60)+" min")
        return tracklabels


    def perform_track(self, labels):
        """ Do tracking with laptrack module """
        ol = OverLapTrack(
                track_cost_cutoff=self.cost_cutoff,
                track_dist_metric_coefs=(1.0, 0.0, -1.0, 0.0, 0.0),
                #gap_closing_dist_metric_coefs=(1.0, -1.0, 0.0, 0.0, 0.0),
                #gap_closing_max_frame_count=1,
                gap_closing_cost_cutoff=False,
                merging_cost_cutoff=self.merging_cost,
                merging_dist_metric_coefs=(1.0, 0.0, 0.0, -1.0, 0.0),
                splitting_cost_cutoff=self.splitting_cost,
                splitting_dist_metric_coefs=(1.0, 0.0, 0.0, 0.0, -1.0),
        )
        
        track_df, split_df, merge_df = ol.predict_overlap_dataframe(labels)
        track_df = track_df.reset_index()
        return track_df, split_df, merge_df

    def track_overlaps(self, labels):
        """ Track all movie with laptrack overlap method """
        ut.napari_info("Starting tracking with laptrack Overlap...")
        return self.perform_track( labels )


    def inspect_oneframe(self, graph, trackdf):
        for track in np.unique(trackdf["track_id"]):
            tr = trackdf[trackdf["track_id"] == track]
            ## track is only on one frame, suspect
            if len(np.unique(tr["frame"])) == 1:
                # trackid + 1 as trackid starts as 0
                pos = (tr.iloc[0]["frame"], int(tr.iloc[0]["centroid-0"]), int(tr.iloc[0]["centroid-1"]))
                self.epicure.inspecting.add_event( pos, track+1, "tracking" )
                if self.track.suggesting:
                    if track in graph.keys():
                        sisters = []
                        refval = graph[track][0]
                        for key, val in graph.items():
                            if val[0] == refval:
                                sisters.append( key )
                        if len(sisters) == 2:
                            for sis in sisters:
                                self.epicure.add_suggestion( sis+1, refval+1 )





import napari
from napari import current_viewer
from magicgui import magicgui
from napari.utils.history import get_save_history, update_save_history 
import pathlib
import epicure.Utils as ut
import numpy as np
import os
import pandas as pd
from epicure.laptrack_centroids import LaptrackCentroids
from skimage.measure import regionprops_table

from epicure.epicuring import EpiCure
from epicure.tracking import Tracking


"""
    Concatenate temporally two epicured movies (the intensity movie and the labels).
    The last frame of the first movie should be the same as the first frame of the second movie.
"""
    
def merge_epicures( first_movie, first_labels, second_movie, second_labels, fullname="fullmovie.tif" ):
    """ Do the concatenation """

    ## open the two intensity movies
    ut.show_info("Reading and concatenating the intensity movies...")
    
    print("Reading first movie and EpiCure files")
    ## open the first movie and epicure data
    first_epicure = EpiCure()
    first_epicure.viewer = napari.Viewer(show=False)
    first_epicure.load_movie( first_movie )
    first_epicure.go_epicure( "epics", first_labels )

    ## open the second movie
    print("Reading second movie and EpiCure files")
    second_epicure = EpiCure()
    second_epicure.viewer = napari.Viewer(show=False)
    second_epicure.load_movie( second_movie )
    second_epicure.go_epicure( "epics", second_labels )
    
    ## check that the last frame of first movie and first frame of second are the same
    if np.sum(first_epicure.img[first_epicure.nframes-1] - second_epicure.img[0]) != 0:
        ut.show_error("Error: the last frame of first movie is not the same of the first frame of the second movie, I cannot concatenate them.")
        return
    
    ## create the full movie
    full_mov = np.concatenate( (first_epicure.img, second_epicure.img[1:]), axis=0 )
    full_movie_name = os.path.join(os.path.dirname(first_epicure.imgpath), fullname) 
    ut.writeTif( full_mov, full_movie_name, first_epicure.epi_metadata["ScaleXY"], first_epicure.img.dtype, what="Full movie") 
    print("Full movie created")

    ###### Read and merge the EpiCure data and merge the movie labels
    ut.show_info("Reading and concatenating the Epicured results...")

    ### Create new EpiCure instance with the full movie
    epic = EpiCure()
    epic.viewer = napari.current_viewer()
    epic.load_movie(full_movie_name)
    epic.verbose = 1
    epic.set_names( "epics" )
    
    ## initialize groups to first movie groups
    epic.groups = first_epicure.groups
    
    ## initialize tracks to first movie tracks
    epic.tracking = Tracking(epic.viewer, epic)
    epic.tracking.track_data = first_epicure.tracking.track_data

    ## initialize track graph to first movie graph
    if first_epicure.tracking.graph is not None:
        epic.tracking.graph = first_epicure.tracking.graph
    else:
        epic.tracking.graph = {}

    ### Now, work on the labels
        
    ## track between last and first frame to associate the labels (can be a little different if corrections were done)
    parent_labels, labels = ut.match_labels( first_epicure.seg[first_epicure.nframes-1], second_epicure.seg[0] )
        
    #### Update all the labels in the second movie to match with the first movie
    nextlabels = first_epicure.get_free_labels(20)  ## prepare a list of unused labels
    used_labels = first_epicure.get_labels()
    used_labels = used_labels + nextlabels
    seconds = np.copy(second_epicure.seg)
    seconds_lab = np.unique(seconds)
    #second_tracks = np.copy(second_epicure.tracking.track_data)
    if second_epicure.tracking.graph is None:
        second_epicure.tracking.graph = {}
    second_graph = second_epicure.tracking.graph.copy()
    ## shift all values and keys in the graph, to be sure it's never the same
    shift = np.max(second_epicure.seg)+1
    keys = list(second_graph.keys())
    for key in keys:
        val = second_graph.pop(key)
        shift_val = []
        for v in val:
            shift_val.append(v+shift)
        second_graph[key+shift] = shift_val

    for lab in seconds_lab:
        if lab > 0:
            newlab = None
            if lab in labels:
                newlab = parent_labels[labels.index(lab)]
                #track_indexes = second_epicure.tracking.get_track_indexes( lab )
                ## Label from second movie has been associated to one from first movie
                if newlab is not None:
                    ### If the label is in one group in first or second movie, update its info in full movie
                    fullmovie_group( epic, newlab, lab, second_epicure )
                    np.place(second_epicure.seg, seconds==lab, newlab)
                    #second_tracks[track_indexes, 0] = newlab
            if newlab is None:
                ## Label is a new cell (not present in first movie)
                nextlabel = nextlabels.pop(0)
                np.place(second_epicure.seg, seconds==lab, nextlabel)
                newlab = nextlabel
                #second_tracks[track_indexes, 0] = nextlabel
                ### add group information if there is one
                if second_epicure.groups is not None:
                    second = second_epicure.find_group( lab )
                    if second is not None:
                        epic.cells_ingroup( [nextlabel], second )
                if len(nextlabels) <= 0:
                    ## the list of unused labels has been completly used, regenerates
                    nextlabels = ut.get_free_labels( used_labels, 20 )
                    used_labels = used_labels + nextlabels

            # add division or merge if lab is in second movie graph
            if (shift+lab) in second_graph.keys():
                second_graph[newlab] = second_graph[shift+lab]
            for key, vals in second_graph.items():
                new_vals = []
                for val in vals:
                    if (shift+lab) == val:
                        new_vals.append( newlab )
                    else:
                        new_vals.append( val )
                second_graph[key] = new_vals

    ### merge the two graphs
    for key, vals in second_graph.items():
        if key not in epic.tracking.graph.keys():
            epic.tracking.graph[key] = vals
        else:
            print("Key "+str(key)+" present in both graph, something might be wrong ")

    ### Ok, save the results
    full_lab = np.concatenate( (first_epicure.seg, second_epicure.seg[1:]), axis=0 )
    epic.seg = full_lab
    epic.save_epicures()
    print("Movie and EpiCure files merged; Suspects/Events (if any) are not merged, use inspect tracks on merged movie to generate them")

def fullmovie_group( epic, first_label, second_label, second_epicure ):
    """ Check if second_label is in a group, and add it to full movie groups if relevant """
    first_group = epic.find_group( first_label )
    second = None
    if second_epicure.groups is not None:
        second = second_epicure.find_group( second_label )
    if (first_group is not None) and (second is not None):
        ### label is present in the two movie groups
        if first_group != second:
            print("Label "+str(first_label)+" classified in group "+(first_group)+" in the first movie and in group "+second+" in the second movie")
            print("Keep only the first movie group: "+first_group)
    else:
        if second is not None:
            epic.cells_ingroup( [first_label], second )

def concatenate_movies():
    hist = get_save_history()
    cdir = hist[0]
    viewer = current_viewer()

    def choose_first_movie():
        """ First movie is chosen, suggest default labels file """
        first = get_files.first_movie.value
        imgname, imgdir, out = ut.extract_names( first, "epics", mkdir=False )
        get_files.first_labels.value = pathlib.Path(out)
        labname = ut.suggest_segfile(out, imgname)
        if labname is not None:
            get_files.first_labels.value = pathlib.Path(labname)
    
    def choose_second_movie():
        """ Second movie is chosen, suggest default labels file """
        second = get_files.second_movie.value
        imgname, imgdir, out = ut.extract_names( second, "epics", mkdir=False )
        get_files.second_labels.value = pathlib.Path(out)
        labname = ut.suggest_segfile(out, imgname)
        if labname is not None:
            get_files.second_labels.value = pathlib.Path(labname)
    
    @magicgui(call_button="Concatenate",)
    def get_files(
            first_movie = pathlib.Path(cdir),
            first_labels = pathlib.Path(cdir),
            second_movie = pathlib.Path(cdir),
            second_labels = pathlib.Path(cdir),
            merged_movie_name = "fullmovie.tif"
            ):
        merge_epicures( first_movie, first_labels, second_movie, second_labels, merged_movie_name)
        return
    
    get_files.first_movie.changed.connect(choose_first_movie)
    get_files.second_movie.changed.connect(choose_second_movie)
    return get_files
